# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7600 - Module for testing async pipelining.
"""

import unittest
import datetime
import decimal

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):

    async def test_7600(self):
        "7600 - test execute() and fetchall()."
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:1)", [1]
        )
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:val)", dict(val=2)
        )
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol"
        )
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [(1,), (2,)])

    async def test_7601(self):
        "7601 - test executemany()"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)", [(2,), (3,)]
        )
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:data)",
            [{"data": 4}, {"data": 5}],
        )
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol"
        )
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [(2,), (3,), (4,), (5,)])

    async def test_7602(self):
        "7602 - test fetchall() with arraysize"
        data = [(1,), (2,), (3,), (4,)]
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:value)",
            [{"value": i} for i, in data],
        )
        pipeline.add_commit()
        arraysize = 1
        op = pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
            arraysize=arraysize,
        )
        self.assertEqual(op.arraysize, arraysize)
        arraysize = len(data)
        op = pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
            arraysize=arraysize,
        )
        self.assertEqual(op.arraysize, arraysize)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, data)
        self.assertEqual(results[-2].rows, data)

    async def test_7603(self):
        "7603 - test fetchall() with rowfactory"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test_7003')
            """
        )
        pipeline.add_commit()

        def rowfactory(*row):
            column_names = ["INTCOL", "STRINGCOL1"]
            return dict(zip(column_names, row))

        pipeline.add_fetchall(
            "select IntCol, StringCol1 from TestTempTable",
            rowfactory=rowfactory,
        )
        results = await self.conn.run_pipeline(pipeline)
        expected_value = [{"INTCOL": 1, "STRINGCOL1": "test_7003"}]
        self.assertEqual(results[-1].rows, expected_value)

    async def test_7604(self):
        "7604 - test fetchone()"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)", [(9,), (10,)]
        )
        pipeline.add_commit()
        pipeline.add_fetchone(
            "select IntCol from TestTempTable order by IntCol"
        )
        pipeline.add_fetchone("select :1 from dual", [23])
        pipeline.add_fetchone("select :val from dual", {"val": 5})
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-3].rows, [(9,)])
        self.assertEqual(results[-2].rows, [(23,)])
        self.assertEqual(results[-1].rows, [(5,)])

    async def test_7605(self):
        "7605 - test fetchone() with rowfactory"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int, :str)
            """,
            [{"int": 3, "str": "Mac"}, {"int": 4, "str": "Doc"}],
        )
        pipeline.add_commit()

        def rowfactory(*row):
            column_names = ["INT", "STRING"]
            return dict(zip(column_names, row))

        op = pipeline.add_fetchone(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
            rowfactory=rowfactory,
        )
        self.assertEqual(op.rowfactory, rowfactory)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [{"INT": 3, "STRING": "Mac"}])

    async def test_7606(self):
        "7606 - test fetchmany()"
        data = [(i,) for i in range(10)]
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        pipeline.add_commit()
        pipeline.add_fetchmany(
            "select IntCol from TestTempTable order by IntCol"
        )
        pipeline.add_fetchmany("select :1 from dual", [1099])
        pipeline.add_fetchmany("select :val from dual", {"val": 366})
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-3].rows, data)
        self.assertEqual(results[-2].rows, [(1099,)])
        self.assertEqual(results[-1].rows, [(366,)])

    async def test_7607(self):
        "7607 - test fetchmany() with num_rows"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        data = [(i,) for i in range(10)]
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        pipeline.add_commit()
        num_rows = 7
        op = pipeline.add_fetchmany(
            "select IntCol from TestTempTable order by IntCol",
            num_rows=num_rows,
        )
        self.assertEqual(op.num_rows, num_rows)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, data[:num_rows])

    async def test_7608(self):
        "7608 - test fetchmany() with rowfactory and num_rows"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int, :str)
            """,
            [{"int": 29, "str": "Feb"}, {"int": 4, "str": "Monday"}],
        )
        pipeline.add_commit()

        def rowfactory(*row):
            column_names = ["INT", "STRING"]
            return dict(zip(column_names, row))

        num_rows = 2
        op = pipeline.add_fetchmany(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
            num_rows=num_rows,
            rowfactory=rowfactory,
        )
        self.assertEqual(op.num_rows, num_rows)
        self.assertEqual(op.rowfactory, rowfactory)
        expected_value = [
            {"INT": 4, "STRING": "Monday"},
            {"INT": 29, "STRING": "Feb"},
        ]
        num_rows = 1
        op = pipeline.add_fetchmany(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
            rowfactory=rowfactory,
            num_rows=num_rows,
        )
        self.assertEqual(op.num_rows, num_rows)
        self.assertEqual(op.rowfactory, rowfactory)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-2].rows, expected_value)
        self.assertEqual(results[-1].rows, [{"INT": 4, "STRING": "Monday"}])

    async def test_7609(self):
        "7609 - test callfunc(), return_value and return_type"
        pipeline = oracledb.create_pipeline()
        pipeline.add_callfunc("func_Test", oracledb.DB_TYPE_NUMBER, ("Yes", 7))
        kwargs = {"a_String": "Keyword", "a_ExtraAmount": 12}
        pipeline.add_callfunc(
            "func_Test", oracledb.DB_TYPE_NUMBER, keyword_parameters=kwargs
        )

        # paramters and keyword parameters
        kwargs = {"a_ExtraAmount": 25}
        func_name = "func_Test"
        op = pipeline.add_callfunc(
            func_name, oracledb.DB_TYPE_NUMBER, ["Mixed"], kwargs
        )
        self.assertEqual(op.name, func_name)
        self.assertEqual(op.return_type, oracledb.DB_TYPE_NUMBER)
        self.assertIsNone(op.statement)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[0].return_value, 10)
        self.assertEqual(results[1].return_value, 19)
        self.assertEqual(results[2].return_value, 30)

    async def test_7610(self):
        "7610 - test callproc() with parameters"
        pipeline = oracledb.create_pipeline()
        var = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        proc_name = "proc_Test"
        params = ("hi", 5, var)
        op = pipeline.add_callproc(proc_name, params)
        self.assertEqual(op.name, proc_name)
        self.assertEqual(op.parameters, params)
        self.assertIsNone(op.keyword_parameters)
        self.assertIsNone(op.statement)
        self.assertEqual(op.arraysize, 0)
        self.assertEqual(op.num_rows, 0)
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(var.getvalue(), 2)

    async def test_7611(self):
        "7611 - test callproc() with keyword_parameters"
        in_out_value = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        in_out_value.setvalue(0, 7)
        out_value = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        params = []
        kwargs = dict(
            a_InValue="Peace", a_InOutValue=in_out_value, a_OutValue=out_value
        )
        pipeline = oracledb.create_pipeline()
        op = pipeline.add_callproc("proc_Test", params, kwargs)
        self.assertEqual(op.parameters, params)
        self.assertEqual(op.keyword_parameters, kwargs)
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(in_out_value.getvalue(), 35)
        self.assertEqual(out_value.getvalue(), 5)

    async def test_7612(self):
        "7612 - test callproc() with parameters and keyword_parameters"
        in_out_value = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        in_out_value.setvalue(0, 8)
        out_value = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        params = ["Input_7612"]
        kwargs = dict(a_InOutValue=in_out_value, a_OutValue=out_value)
        pipeline = oracledb.create_pipeline()
        op = pipeline.add_callproc("proc_Test", params, kwargs)
        self.assertEqual(op.parameters, params)
        self.assertEqual(op.keyword_parameters, kwargs)
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(in_out_value.getvalue(), 80)
        self.assertEqual(out_value.getvalue(), 10)

    async def test_7613(self):
        "7613 - test fetchmany() num_rows with 0 and negative values"
        data = [(i,) for i in range(10)]
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        sql = "select IntCol from TestTempTable"
        op = pipeline.add_fetchmany(sql, num_rows=0)
        self.assertEqual(op.statement, sql)
        with self.assertRaises(OverflowError):
            pipeline.add_fetchmany(sql, num_rows=-1)
        with self.assertRaises(OverflowError):
            pipeline.add_fetchmany(sql, num_rows=-10)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [])

    async def test_7614(self):
        "7614 - test add_commit with transaction_in_progress"
        self.assertFalse(self.conn.transaction_in_progress)
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (5)",
        )
        await self.conn.run_pipeline(pipeline)
        self.assertTrue(self.conn.transaction_in_progress)
        pipeline = oracledb.create_pipeline()
        pipeline.add_commit()
        await self.conn.run_pipeline(pipeline)
        self.assertFalse(self.conn.transaction_in_progress)

    async def test_7615(self):
        "7615 - test getting an error in the middle of pipeline"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:1)", [5]
        )
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (9, 'too many values')"
        )
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        results = await self.conn.run_pipeline(
            pipeline, continue_on_error=True
        )
        expected_value = [(5,)]
        self.assertEqual(results[-3].rows, expected_value)
        self.assertEqual(results[-2].error.full_code, "ORA-00913")
        self.assertEqual(results[-1].rows, expected_value)

    async def test_7617(self):
        "7617 - test insert and update the inserted row"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute("insert into TestTempTable (IntCol) values (5)")
        pipeline.add_execute(
            "update TestTempTable set IntCol=25 where IntCol=5"
        )
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [(25,)])

    async def test_7618(self):
        "7618 - test insert and update inserted rows"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)",
            [(i,) for i in range(100)],
        )
        pipeline.add_execute("update TestTempTable set StringCol1 = 'UPD'")
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
        )
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [(i, "UPD") for i in range(100)])

    async def test_7619(self):
        "7619 - test insert many rows twice"
        values1 = [(i,) for i in range(100)]
        values2 = [(i,) for i in range(200, 205)]
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol) values (:1)"
        pipeline.add_executemany(sql, values1)
        pipeline.add_executemany(sql, values2)
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, values1 + values2)

    async def test_7620(self):
        "7620 - test insert and delete value"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)",
            [(i,) for i in range(100)],
        )
        pipeline.add_execute("delete TestTempTable")
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [])

    async def test_7621(self):
        "7621 - test PipelineOp op_type"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable (IntCol) values (:1)",
            [(1,), (2,)],
        )
        pipeline.add_commit()
        sql = "select IntCol from TestTempTable"
        pipeline.add_fetchone(sql)
        pipeline.add_fetchall(sql)
        pipeline.add_fetchmany(sql)
        pipeline.add_callproc("proc_Test", ("hi", 5, 2))
        pipeline.add_callfunc("func_Test", oracledb.DB_TYPE_NUMBER, ("Yes", 7))
        results = await self.conn.run_pipeline(pipeline)
        expected_values = [
            oracledb.PIPELINE_OP_TYPE_EXECUTE,
            oracledb.PIPELINE_OP_TYPE_EXECUTE_MANY,
            oracledb.PIPELINE_OP_TYPE_COMMIT,
            oracledb.PIPELINE_OP_TYPE_FETCH_ONE,
            oracledb.PIPELINE_OP_TYPE_FETCH_ALL,
            oracledb.PIPELINE_OP_TYPE_FETCH_MANY,
            oracledb.PIPELINE_OP_TYPE_CALL_PROC,
            oracledb.PIPELINE_OP_TYPE_CALL_FUNC,
        ]
        for result, expected_value in zip(results, expected_values):
            self.assertEqual(result.operation.op_type, expected_value)

    async def test_7622(self):
        "7622 - test Pipeline, PipelineOp and PipelineOpResult repr()"
        pipeline = oracledb.create_pipeline()
        self.assertEqual(
            repr(pipeline), "<oracledb.Pipeline with 0 operations>"
        )
        op = pipeline.add_commit()
        self.assertEqual(
            repr(pipeline), "<oracledb.Pipeline with 1 operations>"
        )
        self.assertEqual(repr(op), "<oracledb.PipelineOp of type COMMIT>")
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(
            repr(results[0]),
            "<oracledb.PipelineOpResult for operation of type COMMIT>",
        )

    async def test_7623(self):
        "7623 - test getting an error at the beginning of a pipeline"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table NonExistentTable")
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:1)", [5]
        )
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        results = await self.conn.run_pipeline(
            pipeline, continue_on_error=True
        )
        expected_value = [(5,)]
        self.assertEqual(results[0].error.full_code, "ORA-00942")
        self.assertEqual(results[-1].rows, expected_value)

    async def test_7624(self):
        "7624 - test getting an error at the end of pipeline"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:1)", [5]
        )
        pipeline.add_fetchall(
            "select IntCol from TestTempTable order by IntCol",
        )
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:1)", [5]
        )
        results = await self.conn.run_pipeline(
            pipeline, continue_on_error=True
        )
        expected_value = [(5,)]
        self.assertEqual(results[-2].rows, expected_value)
        self.assertEqual(results[-1].error.full_code, "ORA-00001")

    async def test_7625(self):
        "7625 - test pipeline with clobs"
        clob = await self.conn.createlob(oracledb.DB_TYPE_CLOB, "Temp CLOB")
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("delete from TestCLOBs")
        pipeline.add_execute(
            "insert into TestCLOBs (IntCol, CLOBCol) values (1, :1)", ["CLOB"]
        )
        pipeline.add_execute(
            "insert into TestCLOBs (IntCol, CLOBCol) values (2, :1)", [clob]
        )
        pipeline.add_fetchall(
            "select CLOBCol from TestCLOBs order by IntCol",
        )
        results = await self.conn.run_pipeline(pipeline)
        rows = results[-1].rows

        self.assertEqual(
            [await lob.read() for lob, in rows], ["CLOB", "Temp CLOB"]
        )

    async def test_7626(self):
        "7626 - test nested cursors"
        sql = """
            select 'Level 1 String',
                cursor(
                    select 'Level 2 String',
                        cursor(
                            select 'Level3 String' from dual
                        ) from dual
                ) from dual
        """
        pipeline = oracledb.create_pipeline()
        pipeline.add_fetchone(sql)
        pipeline.add_fetchone("select user from dual")
        results = await self.conn.run_pipeline(pipeline)
        rows = results[0].rows

        async def transform_row(r):
            return tuple([await transform_fn(v) for v in r])

        async def transform_fn(v):
            if isinstance(v, oracledb.AsyncCursor):
                return [await transform_row(r) async for r in v]
            return v

        rows = [await transform_row(r) async for r in rows[0][1]]
        self.assertEqual(rows, [("Level 2 String", [("Level3 String",)])])

    async def test_7627(self):
        "7627 - test executemany with number of iterations"
        num_iterations = 4
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("delete from TestAllTypes")
        pipeline.add_executemany(
            "insert into TestAllTypes (NumberValue) values (1)", num_iterations
        )
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select NumberValue from TestAllTypes order by NumberValue",
        )
        results = await self.conn.run_pipeline(pipeline)
        expected_value = [(1,) for _ in range(num_iterations)]
        self.assertEqual(results[-1].rows, expected_value)

    async def test_7628(self):
        "7628 - test anonymous PL/SQL"
        var = self.cursor.var(int)
        pipeline = oracledb.create_pipeline()
        sql = "begin :var := :value; end;"
        pipeline.add_execute(sql, [var, 5])
        pipeline.add_execute(sql, [var, 10])
        pipeline.add_execute(sql, [var, 15])
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(var.getvalue(), 15)

    async def test_7629(self):
        "7629 - test executemany() with PL/SQL"
        values = [31, 6, 21, 17, 43]
        out_bind = self.cursor.var(
            oracledb.DB_TYPE_NUMBER, arraysize=len(values)
        )
        data = [(i, f"Test {i}", out_bind) for i in values]
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            """
            begin
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol into :out_bind;
            end;
            """,
            data,
        )
        pipeline.add_commit()
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(out_bind.values, values)

    async def test_7630(self):
        "7630 - test oracledb.defaults.fetch_lobs"
        clob_1_value = "CLOB Data One"
        clob_2_value = "CLOB Data Two"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("delete from TestCLOBs")
        pipeline.add_execute(
            "insert into TestCLOBs (IntCol, CLOBCol) values (1, :1)",
            [clob_1_value],
        )
        clob = await self.conn.createlob(oracledb.DB_TYPE_CLOB, clob_2_value)
        pipeline.add_execute(
            "insert into TestCLOBs (IntCol, CLOBCol) values (2, :1)", [clob]
        )
        pipeline.add_fetchall(
            "select CLOBCol from TestCLOBs order by IntCol",
        )
        with test_env.DefaultsContextManager("fetch_lobs", False):
            res = await self.conn.run_pipeline(pipeline)
            self.assertEqual(
                [res[-1].rows], [[(clob_1_value,), (clob_2_value,)]]
            )

    async def test_7631(self):
        "7631 - test pipeline with lobs > 32K"
        blob_1_data = b"T" * 33000
        blob_2_data = b"B" * 33000
        blob = await self.conn.createlob(oracledb.DB_TYPE_BLOB, blob_1_data)
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("delete from TestBLOBs")
        pipeline.add_execute(
            "insert into TestBLOBs (IntCol, BLOBCol) values (1, :1)", [blob]
        )
        pipeline.add_execute(
            "insert into TestBLOBs (IntCol, BLOBCol) values (2, :1)",
            [blob_2_data],
        )
        pipeline.add_fetchall(
            "select BLOBCol from TestBLOBs order by IntCol",
        )
        res = await self.conn.run_pipeline(pipeline)
        expected_value = [blob_1_data, blob_2_data]
        fetched_value = [await lob.read() for lob, in res[-1].rows]
        self.assertEqual(fetched_value, expected_value)

    async def test_7632(self):
        "7632 - test ref cursor"
        ref_cursor1 = self.conn.cursor()
        ref_cursor2 = self.conn.cursor()
        sql = """
                begin
                    open :pcursor for
                        select IntCol
                        from TestNumbers
                        order by IntCol;
                end;"""
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute(sql, [ref_cursor1])
        pipeline.add_execute(sql, [ref_cursor2])
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(
            await ref_cursor1.fetchall(), await ref_cursor2.fetchall()
        )

    async def test_7633(self):
        "7633 - test add_callproc() with ref cursor"
        values = [(2, None, None, None), (3, None, None, None)]
        ref_cursor = self.conn.cursor()
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_executemany(
            "insert into TestTempTable values (:1, :2, :3, :4)", values
        )
        pipeline.add_callproc("myrefcursorproc", [ref_cursor])
        await self.conn.run_pipeline(pipeline)
        self.assertEqual(await ref_cursor.fetchall(), values)

    async def test_7634(self):
        "7634 - test empty pipeline"
        pipeline = oracledb.create_pipeline()
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results, [])

    async def test_7635(self):
        "7635 - test alter session"
        sql = """
            select value FROM nls_session_parameters
            WHERE parameter = 'NLS_DATE_FORMAT'
        """
        (default_date_format,) = await self.conn.fetchone(sql)
        date = datetime.datetime(2000, 12, 15, 7, 3)
        pipeline = oracledb.create_pipeline()
        pipeline.add_fetchone("select to_char(:1) from dual", [date])
        pipeline.add_execute(
            "alter session set NLS_DATE_FORMAT='YYYY-MM-DD HH24:MI'"
        )
        pipeline.add_fetchone("select to_char(:1) from dual", [date])
        pipeline.add_execute(
            f"alter session set NLS_DATE_FORMAT='{default_date_format}'"
        )
        pipeline.add_fetchone("select to_char(:1) from dual", [date])
        pipeline.add_fetchone(sql)
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[2].rows, [("2000-12-15 07:03",)])
        self.assertEqual(results[4].rows, results[0].rows)
        self.assertEqual(results[-1].rows, [(default_date_format,)])

    async def test_7636(self):
        "7636 - test connection inputtypehandler"

        def input_type_handler(cursor, value, num_elements):
            if isinstance(value, str):
                return cursor.var(
                    oracledb.DB_TYPE_NUMBER,
                    arraysize=num_elements,
                    inconverter=lambda x: int(x),
                )

        self.conn.inputtypehandler = input_type_handler
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            "insert into TestTempTable (IntCol) values (:1)", ["12"]
        )
        pipeline.add_commit()
        pipeline.add_fetchall("select IntCol from TestTempTable")
        results = await self.conn.run_pipeline(pipeline)
        self.assertEqual(results[-1].rows, [(12,)])

    async def test_7637(self):
        "7637 - test oracledb.defaults.fetch_decimals"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute("insert into TestTempTable (IntCol) values (1)")
        pipeline.add_fetchall("select IntCol from TestTempTable")
        with test_env.DefaultsContextManager("fetch_decimals", True):
            res = await self.conn.run_pipeline(pipeline)
        self.assertEqual([res[-1].rows], [[(decimal.Decimal("1"),)]])

    async def test_7638(self):
        "7638 - test oracledb.defaults.arraysize"
        arraysize = 1
        with test_env.DefaultsContextManager("arraysize", arraysize):
            data = [(1,), (2,), (3,), (4,)]
            pipeline = oracledb.create_pipeline()
            pipeline.add_execute("truncate table TestTempTable")
            pipeline.add_executemany(
                "insert into TestTempTable (IntCol) values (:value)",
                [{"value": i} for i, in data],
            )
            pipeline.add_commit()
            op = pipeline.add_fetchall(
                "select IntCol from TestTempTable order by IntCol",
            )
            self.assertEqual(op.arraysize, arraysize)
            new_arraysize = 4
            oracledb.defaults.arraysize = new_arraysize
            op = pipeline.add_fetchall(
                "select IntCol from TestTempTable order by IntCol",
            )
            self.assertEqual(op.arraysize, new_arraysize)
            results = await self.conn.run_pipeline(pipeline)
            self.assertEqual(results[-1].rows, data)
            self.assertEqual(results[-2].rows, data)

    async def test_7639(self):
        "7639 - test autocommit"
        conn1 = await test_env.get_connection_async()
        conn1.autocommit = True
        conn2 = await test_env.get_connection_async()

        pipeline1 = oracledb.create_pipeline()
        pipeline1.add_execute("truncate table TestTempTable")
        pipeline1.add_execute("insert into TestTempTable (IntCol) values (1)")

        pipeline2 = oracledb.create_pipeline()
        pipeline2.add_execute("insert into TestTempTable (IntCol) values (2)")
        pipeline2.add_commit()
        pipeline2.add_fetchall(
            "select IntCol from TestTempTable order by IntCol"
        )

        await conn1.run_pipeline(pipeline1)
        results = await conn2.run_pipeline(pipeline2)
        self.assertEqual(results[-1].rows, [(1,), (2,)])

    async def test_7640(self):
        "7640 - test DML returning"
        out_value = self.cursor.var(str, arraysize=2)
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'Value for first row')
            """
        )
        pipeline.add_execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (2, 'Value for second row')
            """
        )
        pipeline.add_execute(
            """
            update TestTempTable set
                StringCol1 = StringCol1 || ' (Modified)'
            returning StringCol1 into :1
            """,
            [out_value],
        )
        pipeline.add_execute("update TestTempTable set StringCol1 = 'Fixed'")
        pipeline.add_commit()
        pipeline.add_fetchall(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        results = await self.conn.run_pipeline(pipeline)
        expected_data = [(1, "Fixed"), (2, "Fixed")]
        self.assertEqual(results[-1].rows, expected_data)
        self.assertEqual(
            out_value.getvalue(),
            [
                "Value for first row (Modified)",
                "Value for second row (Modified)",
            ],
        )

    async def test_7641(self):
        "7641 - test the columns attribute on results"
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'Value for first row')
            """
        )
        pipeline.add_commit()
        pipeline.add_fetchone("select IntCol, StringCol1 from TestTempTable")
        results = await self.conn.run_pipeline(pipeline)
        self.assertIsNone(results[0].columns)
        self.assertIsNone(results[1].columns)
        self.assertIsNone(results[2].columns)
        names = [i.name for i in results[3].columns]
        self.assertEqual(names, ["INTCOL", "STRINGCOL1"])

    async def test_7642(self):
        "7642 - test the columns attribute on single operation"
        pipeline = oracledb.create_pipeline()
        pipeline.add_fetchone("select user from dual")
        results = await self.conn.run_pipeline(pipeline)
        names = [i.name for i in results[0].columns]
        self.assertEqual(names, ["USER"])

    async def test_7643(self):
        "7643 - test DML returning with error - pipeline error"
        out_value = self.cursor.var(oracledb.DB_TYPE_RAW)
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'Value for first row')
            returning StringCol1 into :1
            """,
            [out_value],
        )
        pipeline.add_commit()
        pipeline.add_fetchone("select user from dual")
        with self.assertRaisesFullCode("ORA-01465"):
            await self.conn.run_pipeline(pipeline)
        await self.cursor.execute("select user from dual")
        (fetched_value,) = await self.cursor.fetchone()
        self.assertEqual(fetched_value, test_env.get_main_user().upper())

    async def test_7644(self):
        "7644 - test DML returning with error - pipeline continue"
        out_value = self.cursor.var(oracledb.DB_TYPE_RAW)
        pipeline = oracledb.create_pipeline()
        pipeline.add_execute("truncate table TestTempTable")
        pipeline.add_execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'Value for first row')
            returning StringCol1 into :1
            """,
            [out_value],
        )
        pipeline.add_commit()
        pipeline.add_fetchone("select user from dual")
        results = await self.conn.run_pipeline(
            pipeline, continue_on_error=True
        )
        self.assertEqual(results[1].error.full_code, "ORA-01465")
        user = test_env.get_main_user().upper()
        self.assertEqual(results[3].rows, [(user,)])
        await self.cursor.execute("select user from dual")
        (fetched_value,) = await self.cursor.fetchone()
        self.assertEqual(fetched_value, test_env.get_main_user().upper())


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7800 - Module for testing AQ with raw queues
"""

import oracledb
import test_env
import threading


class TestCase(test_env.BaseTestCase):
    raw_data = [
        b"sample raw data 1",
        b"sample raw data 2",
        b"sample raw data 3",
        b"sample raw data 4",
        b"sample raw data 5",
        b"sample raw data 6",
    ]

    def __deq_in_thread(self, results):
        with test_env.get_connection() as conn:
            queue = conn.queue("TEST_RAW_QUEUE")
            queue.deqoptions.wait = 10
            props = queue.deqone()
            if props is not None:
                results.append(props.payload)
            conn.commit()

    def __verify_attr(self, obj, attrName, value):
        setattr(obj, attrName, value)
        self.assertEqual(getattr(obj, attrName), value)

    def test_7800(self):
        "7800 - test dequeuing an empty RAW queue"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)

    def test_7801(self):
        "7801 - test enqueuing and dequeuing multiple RAW messages"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        props = self.conn.msgproperties()
        for value in self.raw_data:
            props.payload = value
            queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            value = props.payload
            results.append(value)
        self.conn.commit()
        self.assertEqual(results, self.raw_data)

    def test_7802(self):
        "7802 - test dequeuing with DEQ_REMOVE_NODATA in RAW queue"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[1]
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA
        props = queue.deqone()
        self.assertIsNotNone(props)
        self.assertEqual(props.payload, b"")

    def test_7803(self):
        "7803 - test getting/setting dequeue options attributes"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        options = queue.deqoptions
        self.__verify_attr(options, "condition", "TEST_CONDITION")
        self.__verify_attr(options, "consumername", "TEST_CONSUMERNAME")
        self.__verify_attr(options, "correlation", "TEST_CORRELATION")
        self.__verify_attr(options, "mode", oracledb.DEQ_LOCKED)
        self.__verify_attr(
            options, "navigation", oracledb.DEQ_NEXT_TRANSACTION
        )
        self.__verify_attr(options, "transformation", "TEST_TRANSFORMATION")
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)
        self.__verify_attr(options, "wait", 1287)
        self.__verify_attr(options, "msgid", b"mID")

    def test_7804(self):
        "7804 - test enqueue options attributes RAW queue"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        options = queue.enqoptions
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)

    def test_7805(self):
        "7805 - test waiting for dequeue"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        results = []
        thread = threading.Thread(target=self.__deq_in_thread, args=(results,))
        thread.start()
        value = self.raw_data[0]
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)
        self.conn.commit()
        thread.join()
        self.assertEqual(results, [value])

    def test_7806(self):
        "7806 - test getting/setting message properties attributes"
        props = self.conn.msgproperties()
        self.__verify_attr(props, "correlation", "TEST_CORRELATION")
        self.__verify_attr(props, "delay", 60)
        self.__verify_attr(props, "exceptionq", "TEST_EXCEPTIONQ")
        self.__verify_attr(props, "expiration", 30)
        self.assertEqual(props.attempts, 0)
        self.__verify_attr(props, "priority", 1)
        self.assertEqual(props.state, oracledb.MSG_READY)
        self.assertEqual(props.deliverymode, 0)

    def test_7807(self):
        "7807 - test enqueue visibility option - ENQ_ON_COMMIT"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)
        self.conn.commit()
        props = queue.deqone()
        self.assertIsNotNone(props)

    def test_7808(self):
        "7808 - test enqueue visibility option - ENQ_IMMEDIATE"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        value = props.payload
        results = value
        other_conn.commit()
        self.assertEqual(results, self.raw_data[0])

    def test_7809(self):
        "7809 - test enqueue/dequeue delivery modes identical - buffered"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        value = props.payload
        results = value
        other_conn.commit()
        self.assertEqual(results, self.raw_data[0])

    def test_7810(self):
        "7810 - test enqueue/dequeue delivery modes identical - persistent"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        value = props.payload
        results = value
        other_conn.commit()
        self.assertEqual(results, self.raw_data[0])

    def test_7811(self):
        "7811 - test enqueue/dequeue delivery modes the same"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        value = props.payload
        results = value
        other_conn.commit()
        self.assertEqual(results, self.raw_data[0])

    def test_7812(self):
        "7812 - test enqueue/dequeue delivery modes different"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)

        other_conn = test_env.get_connection()
        queue = other_conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertIsNone(props)

    def test_7813(self):
        "7813 - test error for message with no payload"
        queue = self.conn.queue("TEST_RAW_QUEUE")
        props = self.conn.msgproperties()
        with self.assertRaisesFullCode("DPY-2000"):
            queue.enqone(props)

    def test_7814(self):
        "7814 - verify that the msgid property is returned correctly"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        props = self.conn.msgproperties(payload=value)
        self.assertIsNone(props.msgid)
        queue.enqone(props)
        self.cursor.execute("select msgid from RAW_QUEUE_TAB")
        (actual_msgid,) = self.cursor.fetchone()
        self.assertEqual(props.msgid, actual_msgid)
        props = queue.deqone()
        self.assertEqual(props.msgid, actual_msgid)

    def test_7815(self):
        "7815 - test message props enqtime"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        self.cursor.execute("select current_timestamp from dual")
        (start_date,) = self.cursor.fetchone()
        start_date = start_date.replace(microsecond=0)
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)
        props = queue.deqone()
        self.cursor.execute("select current_timestamp from dual")
        (end_date,) = self.cursor.fetchone()
        end_date = end_date.replace(microsecond=0)
        self.assertTrue(start_date <= props.enqtime <= end_date)

    def test_7816(self):
        "7816 - test message props declared attributes"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        values = dict(
            payload=value,
            correlation="TEST_CORRELATION",
            delay=0,
            exceptionq="PYTHONTEST.TEST_EXCEPTIONQ",
            expiration=15,
            priority=1,
        )
        props = self.conn.msgproperties(**values)
        for attr_name in values:
            self.assertEqual(getattr(props, attr_name), values[attr_name])
        queue.enqone(props)
        self.conn.commit()
        prop = queue.deqone()
        for attr_name in values:
            self.assertEqual(getattr(prop, attr_name), values[attr_name])

    def test_7817(self):
        "7817 - test getting queue attributes"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        self.assertEqual(queue.name, "TEST_RAW_QUEUE")
        self.assertEqual(queue.connection, self.conn)

    def test_7818(self):
        "7818 - test getting write-only attributes"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        for options in (queue.enqoptions, queue.deqoptions):
            with self.assertRaises(AttributeError):
                options.deliverymode

    def test_7819(self):
        "7819 - test deqoption condition with priority"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        priorities = [5, 5, 5, 5, 10, 9, 9, 10, 9]
        for priority in priorities:
            value = self.raw_data[0]
            props = self.conn.msgproperties(payload=value, priority=priority)
            queue.enqone(props)

        queue.deqoptions.condition = "priority = 9"
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            results.append(props.payload)
        self.conn.commit()
        self.assertEqual(len(results), 3)

    def test_7820(self):
        "7820 - test deqoption correlation"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        correlations = [
            "sample",
            "sample correlation",
            "sample",
            "sample",
            "sample correlation",
        ]
        for correlation in correlations:
            value = self.raw_data[0]
            props = self.conn.msgproperties(
                payload=value, correlation=correlation
            )
            queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.correlation = "sample correlation"
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            results.append(props.payload)
        self.conn.commit()
        self.assertEqual(len(results), 2)

    def test_7821(self):
        "7821 - test deqoption msgid"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        value = self.raw_data[0]
        props = self.conn.msgproperties(payload=value)
        queue.enqone(props)
        queue.enqone(props)
        self.conn.commit()
        msgid = props.msgid
        queue.enqone(props)
        self.conn.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.msgid = msgid
        prop = queue.deqone()
        self.conn.commit()
        self.assertEqual(prop.msgid, msgid)

    def test_7822(self):
        "7822 - test payload_type returns the correct value"
        queue = self.conn.queue("TEST_RAW_QUEUE")
        self.assertIsNone(queue.payload_type)

    def test_7823(self):
        "7823 - test deprecated attributes (enqOptions, deqOptions)"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        self.assertEqual(queue.enqOptions, queue.enqoptions)
        self.assertEqual(queue.deqOptions, queue.deqoptions)

    def test_7824(self):
        "7824 - test deprecated AQ methods (enqOne, deqOne)"
        value = b"Test 7823"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        queue.enqOne(self.conn.msgproperties(value))
        props = queue.deqOne()
        self.assertEqual(props.payload, value)

    def test_7825(self):
        "7825 - test wrong payload type"
        queue = self.get_and_clear_queue("TEST_RAW_QUEUE")
        typ = self.conn.gettype("UDT_BOOK")
        obj = typ.newobject()
        props = self.conn.msgproperties(payload=obj)
        with self.assertRaisesFullCode("DPY-2062"):
            queue.enqone(props)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# buffer.pyx
#
# Cython file defining the low-level read and write methods for packed data.
#------------------------------------------------------------------------------

cdef enum:
    NUMBER_AS_TEXT_CHARS = 172
    NUMBER_MAX_DIGITS = 40

cdef class Buffer:

    cdef int _get_int_length_and_sign(self, uint8_t *length,
                                      bint *is_negative,
                                      uint8_t max_length) except -1:
        """
        Returns the length of an integer stored in the buffer. A check is also
        made to ensure the integer does not exceed the maximum length. If the
        is_negative pointer is NULL, negative integers will result in an
        exception being raised.
        """
        cdef const char_type *ptr = self._get_raw(1)
        if ptr[0] & 0x80:
            if is_negative == NULL:
                errors._raise_err(errors.ERR_UNEXPECTED_NEGATIVE_INTEGER)
            is_negative[0] = True
            length[0] = ptr[0] & 0x7f
        else:
            if is_negative != NULL:
                is_negative[0] = False
            length[0] = ptr[0]
        if length[0] > max_length:
            errors._raise_err(errors.ERR_INTEGER_TOO_LARGE, length=length[0],
                              max_length=max_length)

    cdef const char_type* _get_raw(self, ssize_t num_bytes) except NULL:
        """
        Returns a pointer to a buffer containing the requested number of bytes.
        """
        cdef:
            ssize_t num_bytes_left
            const char_type *ptr
        num_bytes_left = self._size - self._pos
        if num_bytes > num_bytes_left:
            errors._raise_err(errors.ERR_UNEXPECTED_END_OF_DATA,
                              num_bytes_wanted=num_bytes,
                              num_bytes_available=num_bytes_left)
        ptr = &self._data[self._pos]
        self._pos += num_bytes
        return ptr

    cdef int _initialize(self, ssize_t max_size = TNS_CHUNK_SIZE) except -1:
        """
        Initialize the buffer with an empty bytearray of the specified size.
        """
        self._max_size = max_size
        self._data_obj = bytearray(max_size)
        self._data_view = self._data_obj
        self._data = <char_type*> self._data_obj

    cdef int _populate_from_bytes(self, bytes data) except -1:
        """
        Initialize the buffer with the data in the specified byte string.
        """
        self._max_size = self._size = len(data)
        self._data_obj = bytearray(data)
        self._data_view = self._data_obj
        self._data = <char_type*> self._data_obj

    cdef int _read_raw_bytes_and_length(self, const char_type **ptr,
                                        ssize_t *num_bytes) except -1:
        """
        Helper function that processes the length (if needed) and then acquires
        the specified number of bytes from the buffer. The base function simply
        uses the length as given.
        """
        ptr[0] = self._get_raw(num_bytes[0])

    cdef int _resize(self, ssize_t new_max_size) except -1:
        """
        Resizes the buffer to the new maximum size, copying the data already
        stored in the buffer first.
        """
        cdef:
            bytearray data_obj
            char_type* data
        data_obj = bytearray(new_max_size)
        data = <char_type*> data_obj
        memcpy(data, self._data, self._max_size)
        self._max_size = new_max_size
        self._data_obj = data_obj
        self._data_view = data_obj
        self._data = data

    cdef int _skip_int(self, uint8_t max_length, bint *is_negative) except -1:
        """
        Skips reading an integer of the specified maximum length from the
        buffer.
        """
        cdef uint8_t length
        self._get_int_length_and_sign(&length, is_negative, max_length)
        self.skip_raw_bytes(length)

    cdef int _write_more_data(self, ssize_t num_bytes_available,
                              ssize_t num_bytes_wanted) except -1:
        """
        Called when the amount of buffer available is less than the amount of
        data requested. By default an error is raised.
        """
        errors._raise_err(errors.ERR_BUFFER_LENGTH_INSUFFICIENT,
                          required_buffer_len=num_bytes_wanted,
                          actual_buffer_len=num_bytes_available)

    cdef int _write_raw_bytes_and_length(self, const char_type *ptr,
                                         ssize_t num_bytes) except -1:
        """
        Helper function that writes the length in the format required before
        writing the bytes.
        """
        cdef ssize_t chunk_len
        if num_bytes <= TNS_MAX_SHORT_LENGTH:
            self.write_uint8(<uint8_t> num_bytes)
            if num_bytes > 0:
                self.write_raw(ptr, num_bytes)
        else:
            self.write_uint8(TNS_LONG_LENGTH_INDICATOR)
            while num_bytes > 0:
                chunk_len = min(num_bytes, TNS_CHUNK_SIZE)
                self.write_ub4(chunk_len)
                num_bytes -= chunk_len
                self.write_raw(ptr, chunk_len)
                ptr += chunk_len
            self.write_ub4(0)

    cdef inline ssize_t bytes_left(self):
        """
        Return the number of bytes remaining in the buffer.
        """
        return self._size - self._pos

    cdef int read_oracle_data(self, OracleMetadata metadata,
                              OracleData* data, bint from_dbobject) except -1:
        """
        Reads Oracle data of the given type from the buffer.
        """
        cdef:
            uint8_t ora_type_num
            const uint8_t* ptr
            ssize_t num_bytes
        self.read_raw_bytes_and_length(&ptr, &num_bytes)
        data.is_null = (ptr == NULL)
        if not data.is_null:
            ora_type_num = metadata.dbtype._ora_type_num
            if ora_type_num == ORA_TYPE_NUM_BINARY_DOUBLE:
                decode_binary_double(ptr, num_bytes, &data.buffer)
            elif ora_type_num == ORA_TYPE_NUM_BINARY_FLOAT:
                decode_binary_float(ptr, num_bytes, &data.buffer)
            elif ora_type_num == ORA_TYPE_NUM_BOOLEAN:
                decode_bool(ptr, num_bytes, &data.buffer)
            elif ora_type_num in (
                ORA_TYPE_NUM_CHAR,
                ORA_TYPE_NUM_LONG,
                ORA_TYPE_NUM_LONG_RAW,
                ORA_TYPE_NUM_RAW,
                ORA_TYPE_NUM_VARCHAR,
            ):
                data.buffer.as_raw_bytes.ptr = ptr
                data.buffer.as_raw_bytes.num_bytes = num_bytes
            elif ora_type_num in (
                ORA_TYPE_NUM_DATE,
                ORA_TYPE_NUM_TIMESTAMP,
                ORA_TYPE_NUM_TIMESTAMP_LTZ,
                ORA_TYPE_NUM_TIMESTAMP_TZ,
            ):
                decode_date(ptr, num_bytes, &data.buffer)
            elif ora_type_num == ORA_TYPE_NUM_INTERVAL_DS:
                decode_interval_ds(ptr, num_bytes, &data.buffer)
            elif ora_type_num == ORA_TYPE_NUM_INTERVAL_YM:
                decode_interval_ym(ptr, num_bytes, &data.buffer)
            elif from_dbobject and ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
                data.buffer.as_integer = \
                        <int32_t> decode_integer(ptr, num_bytes)
            elif ora_type_num in (ORA_TYPE_NUM_NUMBER,
                                  ORA_TYPE_NUM_BINARY_INTEGER):
                decode_number(ptr, num_bytes, &data.buffer)
            else:
                errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED,
                                  name=metadata.dbtype.name)

    cdef object read_bytes(self):
        """
        Read bytes from the buffer and return the corresponding Python object
        representing that value.
        """
        cdef:
            const char_type *ptr
            ssize_t num_bytes
        self.read_raw_bytes_and_length(&ptr, &num_bytes)
        if ptr != NULL:
            return ptr[:num_bytes]

    cdef object read_bytes_with_length(self):
        """
        Reads a length from the buffer and then, if the length is non-zero,
        reads bytes from the buffer and returns it.
        """
        cdef uint32_t num_bytes
        self.read_ub4(&num_bytes)
        if num_bytes > 0:
            return self.read_bytes()

    cdef int read_int32be(self, int32_t *value) except -1:
        """
        Read a signed 32-bit integer in big endian order from the buffer.
        """
        value[0] = <int32_t> decode_uint32be(self._get_raw(4))

    cdef const char_type* read_raw_bytes(self, ssize_t num_bytes) except NULL:
        """
        Returns a pointer to a contiguous buffer containing the specified
        number of bytes found in the buffer.
        """
        return self._get_raw(num_bytes)

    cdef int read_raw_bytes_and_length(self, const char_type **ptr,
                                       ssize_t *num_bytes) except -1:
        """
        Reads bytes from the buffer into a contiguous buffer. The first byte
        read is the number of bytes to read.
        """
        cdef uint8_t length
        self.read_ub1(&length)
        if length == 0 or length == TNS_NULL_LENGTH_INDICATOR:
            ptr[0] = NULL
            num_bytes[0] = 0
        else:
            num_bytes[0] = length
            self._read_raw_bytes_and_length(ptr, num_bytes)

    cdef int read_sb1(self, int8_t *value) except -1:
        """
        Reads a signed 8-bit integer from the buffer.
        """
        cdef const char_type *ptr = self._get_raw(1)
        value[0] = <int8_t> ptr[0]

    cdef int read_sb2(self, int16_t *value) except -1:
        """
        Reads a signed 16-bit integer from the buffer.
        """
        cdef:
            const char_type *ptr
            bint is_negative
            uint8_t length
        self._get_int_length_and_sign(&length, &is_negative, 2)
        if length == 0:
            value[0] = 0
        else:
            ptr = self._get_raw(length)
            value[0] = <int16_t> decode_integer(ptr, length)
            if is_negative:
                value[0] = -value[0]

    cdef int read_sb4(self, int32_t *value) except -1:
        """
        Reads a signed 32-bit integer from the buffer.
        """
        cdef:
            const char_type *ptr
            bint is_negative
            uint8_t length
        self._get_int_length_and_sign(&length, &is_negative, 4)
        if length == 0:
            value[0] = 0
        else:
            ptr = self._get_raw(length)
            value[0] = <int32_t> decode_integer(ptr, length)
            if is_negative:
                value[0] = -value[0]

    cdef int read_sb8(self, int64_t *value) except -1:
        """
        Reads a signed 64-bit integer from the buffer.
        """
        cdef:
            const char_type *ptr
            bint is_negative
            uint8_t length
        self._get_int_length_and_sign(&length, &is_negative, 8)
        if length == 0:
            value[0] = 0
        else:
            ptr = self._get_raw(length)
            value[0] = decode_integer(ptr, length)
            if is_negative:
                value[0] = -value[0]

    cdef bytes read_null_terminated_bytes(self):
        """
        Reads null-terminated bytes from the buffer (including the null
        terminator). It is assumed that the buffer contains the full amount. If
        it does not, the remainder of the buffer is returned instead.
        """
        cdef ssize_t start_pos = self._pos, end_pos = self._pos
        while self._data[end_pos] != 0 and end_pos < self._size:
            end_pos += 1
        self._pos = end_pos + 1
        return self._data[start_pos:self._pos]

    cdef object read_str(self, int csfrm, const char* encoding_errors=NULL):
        """
        Reads bytes from the buffer and decodes them into a string following
        the supplied character set form.
        """
        cdef:
            const char_type *ptr
            ssize_t num_bytes
        self.read_raw_bytes_and_length(&ptr, &num_bytes)
        if ptr != NULL:
            if csfrm == CS_FORM_IMPLICIT:
                return ptr[:num_bytes].decode(ENCODING_UTF8, encoding_errors)
            return ptr[:num_bytes].decode(ENCODING_UTF16, encoding_errors)

    cdef object read_str_with_length(self):
        """
        Reads a length from the buffer and then, if the length is non-zero,
        reads string from the buffer and returns it.
        """
        cdef uint32_t num_bytes
        self.read_ub4(&num_bytes)
        if num_bytes > 0:
            return self.read_str(CS_FORM_IMPLICIT)

    cdef int read_ub1(self, uint8_t *value) except -1:
        """
        Reads an unsigned 8-bit integer from the buffer.
        """
        cdef const char_type *ptr = self._get_raw(1)
        value[0] = ptr[0]

    cdef int read_ub2(self, uint16_t *value) except -1:
        """
        Reads an unsigned 16-bit integer from the buffer.
        """
        cdef:
            const char_type *ptr
            uint8_t length
        self._get_int_length_and_sign(&length, NULL, 2)
        if length == 0:
            value[0] = 0
        else:
            ptr = self._get_raw(length)
            value[0] = <uint16_t> decode_integer(ptr, length)

    cdef int read_ub4(self, uint32_t *value) except -1:
        """
        Reads an unsigned 32-bit integer from the buffer.
        """
        cdef:
            const char_type *ptr
            uint8_t length
        self._get_int_length_and_sign(&length, NULL, 4)
        if length == 0:
            value[0] = 0
        else:
            ptr = self._get_raw(length)
            value[0] = <uint32_t> decode_integer(ptr, length)

    cdef int read_ub8(self, uint64_t *value) except -1:
        """
        Reads an unsigned 64-bit integer from the buffer.
        """
        cdef:
            const char_type *ptr
            uint8_t length
        self._get_int_length_and_sign(&length, NULL, 8)
        if length == 0:
            value[0] = 0
        else:
            ptr = self._get_raw(length)
            value[0] = decode_integer(ptr, length)

    cdef int read_uint16be(self, uint16_t *value) except -1:
        """
        Read a 16-bit integer in big endian order from the buffer.
        """
        value[0] = decode_uint16be(self._get_raw(2))

    cdef int read_uint16le(self, uint16_t *value) except -1:
        """
        Read a 16-bit integer in little endian order from the buffer.
        """
        value[0] = decode_uint16le(self._get_raw(2))

    cdef int read_uint32be(self, uint32_t *value) except -1:
        """
        Read a 32-bit integer in big endian order from the buffer.
        """
        value[0] = decode_uint32be(self._get_raw(4))

    cdef int skip_raw_bytes(self, ssize_t num_bytes) except -1:
        """
        Skip the specified number of bytes in the buffer. In order to avoid
        copying data, the number of bytes left in the packet is determined and
        only that amount is requested.
        """
        cdef ssize_t num_bytes_this_time
        while num_bytes > 0:
            num_bytes_this_time = min(num_bytes, self.bytes_left())
            if num_bytes_this_time == 0:
                num_bytes_this_time = num_bytes
            self._get_raw(num_bytes_this_time)
            num_bytes -= num_bytes_this_time

    cdef inline int skip_sb4(self) except -1:
        """
        Skips a signed 32-bit integer in the buffer.
        """
        cdef bint is_negative
        return self._skip_int(4, &is_negative)

    cdef inline void skip_to(self, ssize_t pos):
        """
        Skips to the specified location in the buffer.
        """
        self._pos = pos

    cdef inline int skip_ub1(self) except -1:
        """
        Skips an unsigned 8-bit integer in the buffer.
        """
        self._get_raw(1)

    cdef inline int skip_ub2(self) except -1:
        """
        Skips an unsigned 16-bit integer in the buffer.
        """
        return self._skip_int(2, NULL)

    cdef inline int skip_ub4(self) except -1:
        """
        Skips an unsigned 32-bit integer in the buffer.
        """
        return self._skip_int(4, NULL)

    cdef inline int skip_ub8(self) except -1:
        """
        Skips an unsigned 64-bit integer in the buffer.
        """
        return self._skip_int(8, NULL)

    cdef int write_binary_double(self, double value,
                                 bint write_length=True) except -1:
        """
        Writes a double value to the buffer in Oracle canonical double floating
        point format.
        """
        cdef:
            uint8_t b0, b1, b2, b3, b4, b5, b6, b7
            uint64_t all_bits
            char_type buf[8]
            uint64_t *ptr
        ptr = <uint64_t*> &value
        all_bits = ptr[0]
        b7 = all_bits & 0xff
        b6 = (all_bits >> 8) & 0xff
        b5 = (all_bits >> 16) & 0xff
        b4 = (all_bits >> 24) & 0xff
        b3 = (all_bits >> 32) & 0xff
        b2 = (all_bits >> 40) & 0xff
        b1 = (all_bits >> 48) & 0xff
        b0 = (all_bits >> 56) & 0xff
        if b0 & 0x80 == 0:
            b0 = b0 | 0x80
        else:
            b0 = ~b0
            b1 = ~b1
            b2 = ~b2
            b3 = ~b3
            b4 = ~b4
            b5 = ~b5
            b6 = ~b6
            b7 = ~b7
        buf[0] = b0
        buf[1] = b1
        buf[2] = b2
        buf[3] = b3
        buf[4] = b4
        buf[5] = b5
        buf[6] = b6
        buf[7] = b7
        if write_length:
            self.write_uint8(8)
        self.write_raw(buf, 8)

    cdef int write_binary_float(self, float value,
                                bint write_length=True) except -1:
        """
        Writes a float value to the buffer in Oracle canonical floating point
        format.
        """
        cdef:
            uint8_t b0, b1, b2, b3
            uint32_t all_bits
            char_type buf[4]
            uint32_t *ptr
        ptr = <uint32_t*> &value
        all_bits = ptr[0]
        b3 = all_bits & 0xff
        b2 = (all_bits >> 8) & 0xff
        b1 = (all_bits >> 16) & 0xff
        b0 = (all_bits >> 24) & 0xff
        if b0 & 0x80 == 0:
            b0 = b0 | 0x80
        else:
            b0 = ~b0
            b1 = ~b1
            b2 = ~b2
            b3 = ~b3
        buf[0] = b0
        buf[1] = b1
        buf[2] = b2
        buf[3] = b3
        if write_length:
            self.write_uint8(4)
        self.write_raw(buf, 4)

    cdef int write_bool(self, bint value) except -1:
        """
        Writes a boolean value to the buffer.
        """
        if value:
            self.write_uint8(2)
            self.write_uint16be(0x0101)
        else:
            self.write_uint16be(0x0100)

    cdef int write_bytes(self, bytes value) except -1:
        """
        Writes the bytes to the buffer directly.
        """
        cdef:
            ssize_t value_len
            char_type *ptr
        cpython.PyBytes_AsStringAndSize(value, <char**> &ptr, &value_len)
        self.write_raw(ptr, value_len)

    cdef int write_bytes_with_length(self, bytes value) except -1:
        """
        Writes the bytes to the buffer after first writing the length.
        """
        cdef:
            ssize_t value_len
            char_type *ptr
        cpython.PyBytes_AsStringAndSize(value, <char**> &ptr, &value_len)
        self._write_raw_bytes_and_length(ptr, value_len)

    cdef int write_interval_ds(self, object value,
                               bint write_length=True) except -1:
        """
        Writes an interval to the buffer in Oracle Interval Day To Second
        format.
        """
        cdef:
            int32_t days, seconds, fseconds
            char_type buf[11]
        days = cydatetime.timedelta_days(value)
        encode_uint32be(buf, days + TNS_DURATION_MID)
        seconds = cydatetime.timedelta_seconds(value)
        buf[4] = (seconds // 3600) + TNS_DURATION_OFFSET
        seconds = seconds % 3600
        buf[5] = (seconds // 60) + TNS_DURATION_OFFSET
        buf[6] = (seconds % 60) + TNS_DURATION_OFFSET
        fseconds = cydatetime.timedelta_microseconds(value) * 1000
        encode_uint32be(&buf[7], fseconds + TNS_DURATION_MID)
        if write_length:
            self.write_uint8(sizeof(buf))
        self.write_raw(buf, sizeof(buf))

    cdef int write_interval_ym(self, object value,
                               bint write_length=True) except -1:
        """
        Writes an interval to the buffer in Oracle Interval Day To Second
        format.
        """
        cdef:
            int32_t years, months
            char_type buf[5]
        years = (<tuple> value)[0]
        months = (<tuple> value)[1]
        encode_uint32be(buf, years + TNS_DURATION_MID)
        buf[4] = months + TNS_DURATION_OFFSET
        if write_length:
            self.write_uint8(sizeof(buf))
        self.write_raw(buf, sizeof(buf))

    cdef int write_oracle_date(self, object value, uint8_t length,
                               bint write_length=True) except -1:
        """
        Writes a date to the buffer in Oracle Date format.
        """
        cdef:
            unsigned int year
            char_type buf[13]
            uint32_t fsecond
        year = cydatetime.PyDateTime_GET_YEAR(value)
        buf[0] = <uint8_t> ((year // 100) + 100)
        buf[1] = <uint8_t> ((year % 100) + 100)
        buf[2] = <uint8_t> cydatetime.PyDateTime_GET_MONTH(value)
        buf[3] = <uint8_t> cydatetime.PyDateTime_GET_DAY(value)
        buf[4] = <uint8_t> cydatetime.PyDateTime_DATE_GET_HOUR(value) + 1
        buf[5] = <uint8_t> cydatetime.PyDateTime_DATE_GET_MINUTE(value) + 1
        buf[6] = <uint8_t> cydatetime.PyDateTime_DATE_GET_SECOND(value) + 1
        if length > 7:
            fsecond = <uint32_t> \
                    cydatetime.PyDateTime_DATE_GET_MICROSECOND(value) * 1000
            if fsecond == 0 and length <= 11:
                length = 7
            else:
                encode_uint32be(&buf[7], fsecond)
        if length > 11:
            buf[11] = TZ_HOUR_OFFSET
            buf[12] = TZ_MINUTE_OFFSET
        if write_length:
            self.write_uint8(length)
        self.write_raw(buf, length)

    cdef int write_oracle_number(self, bytes num_bytes) except -1:
        """
        Writes a number in UTF-8 encoded bytes in Oracle Number format to the
        buffer.
        """
        cdef:
            uint8_t num_digits = 0, digit, num_pairs, pair_num, digits_pos
            bint exponent_is_negative = False, append_sentinel = False
            ssize_t num_bytes_length, exponent_pos, pos = 0
            bint is_negative = False, prepend_zero = False
            uint8_t digits[NUMBER_AS_TEXT_CHARS]
            int16_t decimal_point_index
            int8_t exponent_on_wire
            const char_type *ptr
            int16_t exponent

        # zero length string cannot be converted
        num_bytes_length = len(num_bytes)
        if num_bytes_length == 0:
            errors._raise_err(errors.ERR_NUMBER_STRING_OF_ZERO_LENGTH)
        elif num_bytes_length > NUMBER_AS_TEXT_CHARS:
            errors._raise_err(errors.ERR_NUMBER_STRING_TOO_LONG)

        # check to see if number is negative (first character is '-')
        ptr = num_bytes
        if ptr[0] == b'-':
            is_negative = True
            pos += 1

        # scan for digits until the decimal point or exponent indicator found
        while pos < num_bytes_length:
            if ptr[pos] == b'.' or ptr[pos] == b'e' or ptr[pos] == b'E':
                break
            if ptr[pos] < b'0' or ptr[pos] > b'9':
                errors._raise_err(errors.ERR_INVALID_NUMBER)
            digit = ptr[pos] - <uint8_t> b'0'
            pos += 1
            if digit == 0 and num_digits == 0:
                continue
            digits[num_digits] = digit
            num_digits += 1
        decimal_point_index = num_digits

        # scan for digits following the decimal point, if applicable
        if pos < num_bytes_length and ptr[pos] == b'.':
            pos += 1
            while pos < num_bytes_length:
                if ptr[pos] == b'e' or ptr[pos] == b'E':
                    break
                digit = ptr[pos] - <uint8_t> b'0'
                pos += 1
                if digit == 0 and num_digits == 0:
                    decimal_point_index -= 1
                    continue
                digits[num_digits] = digit
                num_digits += 1

        # handle exponent, if applicable
        if pos < num_bytes_length and (ptr[pos] == b'e' or ptr[pos] == b'E'):
            pos += 1
            if pos < num_bytes_length:
                if ptr[pos] == b'-':
                    exponent_is_negative = True
                    pos += 1
                elif ptr[pos] == b'+':
                    pos += 1
            exponent_pos = pos
            while pos < num_bytes_length:
                if ptr[pos] < b'0' or ptr[pos] > b'9':
                    errors._raise_err(errors.ERR_NUMBER_WITH_INVALID_EXPONENT)
                pos += 1
            if exponent_pos == pos:
                errors._raise_err(errors.ERR_NUMBER_WITH_EMPTY_EXPONENT)
            exponent = <int16_t> int(ptr[exponent_pos:pos])
            if exponent_is_negative:
                exponent = -exponent
            decimal_point_index += exponent

        # if there is anything left in the string, that indicates an invalid
        # number as well
        if pos < num_bytes_length:
            errors._raise_err(errors.ERR_CONTENT_INVALID_AFTER_NUMBER)

        # skip trailing zeros
        while num_digits > 0 and digits[num_digits - 1] == 0:
            num_digits -= 1

        # value must be less than 1e126 and greater than 1e-129; the number of
        # digits also cannot exceed the maximum precision of Oracle numbers
        if num_digits > NUMBER_MAX_DIGITS or decimal_point_index > 126 \
                or decimal_point_index < -129:
            errors._raise_err(errors.ERR_ORACLE_NUMBER_NO_REPR)

        # if the exponent is odd, prepend a zero
        if decimal_point_index % 2 == 1:
            prepend_zero = True
            if num_digits > 0:
                digits[num_digits] = 0
                num_digits += 1
                decimal_point_index += 1

        # determine the number of digit pairs; if the number of digits is odd,
        # append a zero to make the number of digits even
        if num_digits % 2 == 1:
            digits[num_digits] = 0
            num_digits += 1
        num_pairs = num_digits // 2

        # append a sentinel 102 byte for negative numbers if there is room
        if is_negative and num_digits > 0 and num_digits < NUMBER_MAX_DIGITS:
            append_sentinel = True

        # write length of number
        self.write_uint8(num_pairs + 1 + append_sentinel)

        # if the number of digits is zero, the value is itself zero since all
        # leading and trailing zeros are removed from the digits string; this
        # is a special case
        if num_digits == 0:
            self.write_uint8(128)
            return 0

        # write the exponent
        exponent_on_wire = <int8_t> (decimal_point_index / 2) + 192
        if is_negative:
            exponent_on_wire = ~exponent_on_wire
        self.write_uint8(exponent_on_wire)

        # write the mantissa bytes
        digits_pos = 0
        for pair_num in range(num_pairs):
            if pair_num == 0 and prepend_zero:
                digit = digits[digits_pos]
                digits_pos += 1
            else:
                digit = digits[digits_pos] * 10 + digits[digits_pos + 1]
                digits_pos += 2
            if is_negative:
                digit = 101 - digit
            else:
                digit += 1
            self.write_uint8(digit)

        # append 102 byte for negative numbers if the number of digits is less
        # than the maximum allowable
        if append_sentinel:
            self.write_uint8(102)

    cdef int write_raw(self, const char_type *data, ssize_t length) except -1:
        """
        Writes raw bytes of the specified length to the buffer.
        """
        cdef ssize_t bytes_to_write
        while True:
            bytes_to_write = min(self._max_size - self._pos, length)
            if bytes_to_write > 0:
                memcpy(self._data + self._pos, <void*> data, bytes_to_write)
                self._pos += bytes_to_write
            if bytes_to_write == length:
                break
            self._write_more_data(self._max_size - self._pos, length)
            length -= bytes_to_write
            data += bytes_to_write

    cdef int write_sb4(self, int32_t value) except -1:
        """
        Writes a 32-bit signed integer to the buffer in universal format.
        """
        cdef uint8_t sign = 0
        if value < 0:
            value = -value
            sign = 0x80
        if value == 0:
            self.write_uint8(0)
        elif value <= UINT8_MAX:
            self.write_uint8(1 | sign)
            self.write_uint8(<uint8_t> value)
        elif value <= UINT16_MAX:
            self.write_uint8(2 | sign)
            self.write_uint16be(<uint16_t> value)
        else:
            self.write_uint8(4 | sign)
            self.write_uint32be(value)

    cdef int write_str(self, str value) except -1:
        """
        Writes a string to the buffer as UTF-8 encoded bytes.
        """
        self.write_bytes(value.encode())

    cdef int write_uint8(self, uint8_t value) except -1:
        """
        Writes an 8-bit integer to the buffer.
        """
        if self._pos + 1 > self._max_size:
            self._write_more_data(self._max_size - self._pos, 1)
        self._data[self._pos] = value
        self._pos += 1

    cdef int write_uint16be(self, uint16_t value) except -1:
        """
        Writes a 16-bit integer to the buffer in big endian format.
        """
        if self._pos + 2 > self._max_size:
            self._write_more_data(self._max_size - self._pos, 2)
        encode_uint16be(&self._data[self._pos], value)
        self._pos += 2

    cdef int write_uint16le(self, uint16_t value) except -1:
        """
        Writes a 16-bit integer to the buffer in little endian format.
        """
        if self._pos + 2 > self._max_size:
            self._write_more_data(self._max_size - self._pos, 2)
        encode_uint16le(&self._data[self._pos], value)
        self._pos += 2

    cdef int write_uint32be(self, uint32_t value) except -1:
        """
        Writes a 32-bit integer to the buffer in big endian format.
        """
        if self._pos + 4 > self._max_size:
            self._write_more_data(self._max_size - self._pos, 4)
        encode_uint32be(&self._data[self._pos], value)
        self._pos += 4

    cdef int write_uint64be(self, uint64_t value) except -1:
        """
        Writes a 64-bit integer to the buffer in big endian format.
        """
        if self._pos + 8 > self._max_size:
            self._write_more_data(self._max_size - self._pos, 8)
        encode_uint64be(&self._data[self._pos], value)
        self._pos += 8

    cdef int write_ub2(self, uint16_t value) except -1:
        """
        Writes a 16-bit integer to the buffer in universal format.
        """
        if value == 0:
            self.write_uint8(0)
        elif value <= UINT8_MAX:
            self.write_uint8(1)
            self.write_uint8(<uint8_t> value)
        else:
            self.write_uint8(2)
            self.write_uint16be(value)

    cdef int write_ub4(self, uint32_t value) except -1:
        """
        Writes a 32-bit integer to the buffer in universal format.
        """
        if value == 0:
            self.write_uint8(0)
        elif value <= UINT8_MAX:
            self.write_uint8(1)
            self.write_uint8(<uint8_t> value)
        elif value <= UINT16_MAX:
            self.write_uint8(2)
            self.write_uint16be(<uint16_t> value)
        else:
            self.write_uint8(4)
            self.write_uint32be(value)

    cdef int write_ub8(self, uint64_t value) except -1:
        """
        Writes a 64-bit integer to the buffer in universal format.
        """
        if value == 0:
            self.write_uint8(0)
        elif value <= UINT8_MAX:
            self.write_uint8(1)
            self.write_uint8(<uint8_t> value)
        elif value <= UINT16_MAX:
            self.write_uint8(2)
            self.write_uint16be(<uint16_t> value)
        elif value <= UINT32_MAX:
            self.write_uint8(4)
            self.write_uint32be(<uint32_t> value)
        else:
            self.write_uint8(8)
            self.write_uint64be(value)


cdef class GrowableBuffer(Buffer):

    cdef int _reserve_space(self, ssize_t num_bytes) except -1:
        """
        Reserves the requested amount of space in the buffer by moving the
        pointer forward, allocating more space if necessary.
        """
        self._pos += num_bytes
        if self._pos > self._size:
            self._write_more_data(self._size - self._pos + num_bytes,
                                  num_bytes)

    cdef int _write_more_data(self, ssize_t num_bytes_available,
                              ssize_t num_bytes_wanted) except -1:
        """
        Called when the amount of buffer available is less than the amount of
        data requested. The buffer is increased in multiples of TNS_CHUNK_SIZE
        in order to accomodate the number of bytes desired.
        """
        cdef:
            ssize_t num_bytes_needed = num_bytes_wanted - num_bytes_available
            ssize_t new_size
        new_size = (self._max_size + num_bytes_needed + TNS_CHUNK_SIZE - 1) & \
                ~(TNS_CHUNK_SIZE - 1)
        self._resize(new_size)



# -----------------------------------------------------------------------------
# Copyright (c) 2021, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3600 - Module for testing the conversions of outputtype handler.
"""

import datetime
import decimal

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def __test_type_handler(
        self, input_type, output_type, in_value, expected_out_value
    ):
        def type_handler(cursor, metadata):
            return cursor.var(output_type, arraysize=cursor.arraysize)

        self.cursor.outputtypehandler = type_handler
        self.assertEqual(self.cursor.outputtypehandler, type_handler)
        var = self.cursor.var(input_type)
        var.setvalue(0, in_value)
        self.cursor.execute("select :1 from dual", [var])
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(type(fetched_value), type(expected_out_value))
        self.assertEqual(fetched_value, expected_out_value)

    def __test_type_handler_lob(self, lob_type, output_type):
        db_type = getattr(oracledb, lob_type)

        def type_handler(cursor, metadata):
            if metadata.type_code is db_type:
                return cursor.var(output_type, arraysize=cursor.arraysize)

        self.cursor.outputtypehandler = type_handler
        in_value = f"Some {lob_type} data"
        if lob_type == "BLOB":
            in_value = in_value.encode()
        self.cursor.execute(f"delete from Test{lob_type}s")
        self.cursor.execute(
            f"""
            insert into Test{lob_type}s (IntCol, {lob_type}Col)
            values(1, :val)
            """,
            val=in_value,
        )
        self.conn.commit()
        self.cursor.execute(
            f"select {lob_type}Col, IntCol, {lob_type}Col from Test{lob_type}s"
        )
        self.assertEqual(self.cursor.fetchone(), (in_value, 1, in_value))

    def setUp(self):
        super().setUp()
        self.cursor.execute(
            """
            ALTER SESSION SET NLS_DATE_FORMAT = 'YYYY-MM-DD HH24:MI:SS'
            NLS_TIMESTAMP_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF6'
            NLS_TIMESTAMP_TZ_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF6'
            time_zone='Europe/London'
            """
        )

    def test_3600(self):
        "3600 - output type handler: from VARCHAR to NUMBER"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER, "31.5", 31.5
        )

    def test_3601(self):
        "3601 - output type handler: from CHAR to NUMBER"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_NUMBER, "31.5", 31.5
        )

    def test_3602(self):
        "3602 - output type handler: from LONG to NUMBER"
        self.__test_type_handler(
            oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_NUMBER, "31.5", 31.5
        )

    def test_3603(self):
        "3603 - test output type handler: from INTEGER to NUMBER"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_INTEGER, oracledb.DB_TYPE_NUMBER, 31, 31
        )

    def test_3604(self):
        "3604 - output type handler: from VARCHAR to INTEGER"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_BINARY_INTEGER,
            "31.5",
            31,
        )

    def test_3605(self):
        "3605 - output type handler: from CHAR to INTEGER"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_BINARY_INTEGER, "31.5", 31
        )

    def test_3606(self):
        "3606 - output type handler: from LONG to INTEGER"
        self.__test_type_handler(
            oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_BINARY_INTEGER, "31.5", 31
        )

    def test_3607(self):
        "3607 - output type handler: from NUMBER to INTEGER"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_BINARY_INTEGER, 31.5, 31
        )

    def test_3608(self):
        "3608 - output type handler: from DOUBLE to INTEGER"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_DOUBLE,
            oracledb.DB_TYPE_BINARY_INTEGER,
            31.5,
            31,
        )

    def test_3609(self):
        "3609 - output type handler: from FLOAT to INTEGER"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_FLOAT,
            oracledb.DB_TYPE_BINARY_INTEGER,
            31.5,
            31,
        )

    def test_3610(self):
        "3610 - output type handler: from DATE to VARCHAR"
        in_val = datetime.date(2021, 2, 1)
        out_val = "2021-02-01 00:00:00"
        self.__test_type_handler(
            oracledb.DB_TYPE_DATE, oracledb.DB_TYPE_VARCHAR, in_val, out_val
        )

    def test_3611(self):
        "3611 - output type handler: from DATE to CHAR"
        in_val = datetime.date(2021, 2, 1)
        out_val = "2021-02-01 00:00:00"
        self.__test_type_handler(
            oracledb.DB_TYPE_DATE, oracledb.DB_TYPE_CHAR, in_val, out_val
        )

    def test_3612(self):
        "3612 - output type handler: from DATE to LONG"
        in_val = datetime.date(2021, 2, 1)
        out_val = "2021-02-01 00:00:00"
        self.__test_type_handler(
            oracledb.DB_TYPE_DATE, oracledb.DB_TYPE_LONG, in_val, out_val
        )

    def test_3613(self):
        "3613 - output type handler: from NUMBER to VARCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_VARCHAR, 31.5, "31.5"
        )
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_VARCHAR, 0, "0"
        )

    def test_3614(self):
        "3614 - output type handler: from NUMBER to CHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_CHAR, 31.5, "31.5"
        )
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_CHAR, 0, "0"
        )

    def test_3615(self):
        "3615 - output type handler: from NUMBER to LONG"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_LONG, 31.5, "31.5"
        )

    def test_3616(self):
        "3616 - output type handler: from INTERVAL to VARCHAR"
        in_val = datetime.timedelta(
            days=-1, seconds=86314, microseconds=431152
        )
        if test_env.get_is_thin():
            out_val = str(in_val)
        else:
            out_val = "-000000001 23:58:34.431152000"
        self.__test_type_handler(
            oracledb.DB_TYPE_INTERVAL_DS,
            oracledb.DB_TYPE_VARCHAR,
            in_val,
            out_val,
        )

    def test_3617(self):
        "3617 - output type handler: from INTERVAL to CHAR"
        in_val = datetime.timedelta(
            days=-1, seconds=86314, microseconds=431152
        )
        if test_env.get_is_thin():
            out_val = str(in_val)
        else:
            out_val = "-000000001 23:58:34.431152000"
        self.__test_type_handler(
            oracledb.DB_TYPE_INTERVAL_DS,
            oracledb.DB_TYPE_CHAR,
            in_val,
            out_val,
        )

    def test_3618(self):
        "3618 - output type handler: from INTERVAL to LONG"
        in_val = datetime.timedelta(
            days=-1, seconds=86314, microseconds=431152
        )
        if test_env.get_is_thin():
            out_val = str(in_val)
        else:
            out_val = "-000000001 23:58:34.431152000"
        self.__test_type_handler(
            oracledb.DB_TYPE_INTERVAL_DS,
            oracledb.DB_TYPE_LONG,
            in_val,
            out_val,
        )

    def test_3619(self):
        "3619 - output type handler: from TIMESTAMP to VARCHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP,
            oracledb.DB_TYPE_VARCHAR,
            in_val,
            str(in_val),
        )

    def test_3620(self):
        "3620 - output type handler: from TIMESTAMP to CHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP,
            oracledb.DB_TYPE_CHAR,
            in_val,
            str(in_val),
        )

    def test_3621(self):
        "3621 - output type handler: from TIMESTAMP to LONG"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP,
            oracledb.DB_TYPE_LONG,
            in_val,
            str(in_val),
        )

    def test_3622(self):
        "3622 - output type handler: from TIMESTAMP_TZ to VARCHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_TZ,
            oracledb.DB_TYPE_VARCHAR,
            in_val,
            str(in_val),
        )

    def test_3623(self):
        "3623 - output type handler: from TIMESTAMP_TZ to CHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_TZ,
            oracledb.DB_TYPE_CHAR,
            in_val,
            str(in_val),
        )

    def test_3624(self):
        "3624 - output type handler: from TIMESTAMP_TZ to LONG"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_TZ,
            oracledb.DB_TYPE_LONG,
            in_val,
            str(in_val),
        )

    def test_3625(self):
        "3625 - output type handler: from TIMESTAMP_LTZ to VARCHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_LTZ,
            oracledb.DB_TYPE_VARCHAR,
            in_val,
            str(in_val),
        )

    def test_3626(self):
        "3626 - output type handler: from TIMESTAMP_LTZ to CHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_LTZ,
            oracledb.DB_TYPE_CHAR,
            in_val,
            str(in_val),
        )

    def test_3627(self):
        "3627 - output type handler: from TIMESTAMP_LTZ to LONG"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_LTZ,
            oracledb.DB_TYPE_LONG,
            in_val,
            str(in_val),
        )

    def test_3628(self):
        "3628 - output type handler: from INTEGER to VARCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_INTEGER, oracledb.DB_TYPE_VARCHAR, 31, "31"
        )

    def test_3629(self):
        "3629 - output type handler: from INTEGER to CHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_INTEGER, oracledb.DB_TYPE_CHAR, 31, "31"
        )

    def test_3630(self):
        "3630 - output type handler: from INTEGER to LONG"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_INTEGER, oracledb.DB_TYPE_LONG, 31, "31"
        )

    def test_3631(self):
        "3631 - output type handler: from NUMBER to DOUBLE"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_BINARY_DOUBLE, 31.5, 31.5
        )

    def test_3632(self):
        "3632 - output type handler: from FLOAT to DOUBLE"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_FLOAT,
            oracledb.DB_TYPE_BINARY_DOUBLE,
            31.5,
            31.5,
        )

    def test_3633(self):
        "3633 - output type handler: from VARCHAR to DOUBLE"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_BINARY_DOUBLE,
            "31.5",
            31.5,
        )

    def test_3634(self):
        "3634 - output type handler: from CHAR to DOUBLE"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_BINARY_DOUBLE, "31.5", 31.5
        )

    def test_3635(self):
        "3635 - output type handler: from LONG to DOUBLE"
        self.__test_type_handler(
            oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_BINARY_DOUBLE, "31.5", 31.5
        )

    def test_3636(self):
        "3636 - output type handler: from NUMBER to FLOAT"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_BINARY_FLOAT, 31.5, 31.5
        )

    def test_3637(self):
        "3637 - output type handler: from DOUBLE to FLOAT"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_DOUBLE,
            oracledb.DB_TYPE_BINARY_FLOAT,
            31.5,
            31.5,
        )

    def test_3638(self):
        "3638 - output type handler: from VARCHAR to FLOAT"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_BINARY_FLOAT,
            "31.5",
            31.5,
        )

    def test_3639(self):
        "3639 - output type handler: from CHAR to FLOAT"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_BINARY_FLOAT, "31.5", 31.5
        )

    def test_3640(self):
        "3640 - output type handler: from LONG to FLOAT"
        self.__test_type_handler(
            oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_BINARY_FLOAT, "31.5", 31.5
        )

    def test_3641(self):
        "3641 - output type handler: from VARCHAR to CHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_CHAR, "31.5", "31.5"
        )

    def test_3642(self):
        "3642 - output type handler: from VARCHAR to LONG"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_LONG, "31.5", "31.5"
        )

    def test_3643(self):
        "3643 - output type handler: from LONG to VARCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_VARCHAR, "31.5", "31.5"
        )

    def test_3644(self):
        "3644 - output type handler: from LONG to CHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_CHAR, "31.5", "31.5"
        )

    def test_3645(self):
        "3645 - output type handler: from CHAR to VARCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_VARCHAR, "31.5", "31.5"
        )

    def test_3646(self):
        "3646 - output type handler: from CHAR to LONG"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_LONG, "31.5", "31.5"
        )

    def test_3647(self):
        "3647 - output type handler: from TIMESTAMP to TIMESTAMP_TZ"
        val = datetime.datetime(2002, 12, 17, 0, 0, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP, oracledb.DB_TYPE_TIMESTAMP_TZ, val, val
        )

    def test_3648(self):
        "3648 - output type handler: from TIMESTAMP to TIMESTAMP_LTZ"
        val = datetime.datetime(2002, 12, 17, 0, 0, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP,
            oracledb.DB_TYPE_TIMESTAMP_LTZ,
            val,
            val,
        )

    def test_3649(self):
        "3649 - output type handler: from TIMESTAMP_TZ to TIMESTAMP"
        val = datetime.datetime(2002, 12, 17, 0, 0, 16, 400000)
        self.__test_type_handler(
            oracledb.DB_TYPE_TIMESTAMP_TZ, oracledb.DB_TYPE_TIMESTAMP, val, val
        )

    def test_3650(self):
        "3650 - output type handler: from NUMBER to DATE is invalid"
        with self.assertRaisesFullCode("DPY-4007", "ORA-00932"):
            self.__test_type_handler(
                oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_DATE, 3, 3
            )

    def test_3651(self):
        "3651 - output type handler: from CLOB to CHAR"
        val = "Some Clob String"
        self.__test_type_handler(
            oracledb.DB_TYPE_CLOB, oracledb.DB_TYPE_CHAR, val, val
        )

    def test_3652(self):
        "3652 - output type handler: from CLOB to VARCHAR"
        val = "Some Clob String"
        self.__test_type_handler(
            oracledb.DB_TYPE_CLOB, oracledb.DB_TYPE_VARCHAR, val, val
        )

    def test_3653(self):
        "3653 - output type handler: from CLOB to LONG"
        val = "Some Clob String"
        self.__test_type_handler(
            oracledb.DB_TYPE_CLOB, oracledb.DB_TYPE_LONG, val, val
        )

    def test_3654(self):
        "3654 - output type handler: from BLOB to RAW"
        val = b"Some binary data"
        self.__test_type_handler(
            oracledb.DB_TYPE_BLOB, oracledb.DB_TYPE_RAW, val, val
        )

    def test_3655(self):
        "3655 - output type handler: from BLOB to LONGRAW"
        val = b"Some binary data"
        self.__test_type_handler(
            oracledb.DB_TYPE_BLOB, oracledb.DB_TYPE_LONG_RAW, val, val
        )

    def test_3656(self):
        "3656 - output type handler: from permanent BLOBs to LONG_RAW"
        self.__test_type_handler_lob("BLOB", oracledb.DB_TYPE_LONG_RAW)

    def test_3657(self):
        "3657 - output type handler: from permanent BLOBs to RAW"
        self.__test_type_handler_lob("BLOB", oracledb.DB_TYPE_RAW)

    def test_3658(self):
        "3658 - output type handler: from permanent CLOBs to VARCHAR"
        self.__test_type_handler_lob("CLOB", oracledb.DB_TYPE_VARCHAR)

    def test_3659(self):
        "3659 - output type handler: from permanent CLOBs to CHAR"
        self.__test_type_handler_lob("CLOB", oracledb.DB_TYPE_CHAR)

    def test_3660(self):
        "3660 - output type handler: from permanent CLOBs to LONG"
        self.__test_type_handler_lob("CLOB", oracledb.DB_TYPE_LONG)

    def test_3661(self):
        "3661 - output type handler: from NCLOB to CHAR"
        val = "Some nclob data"
        self.__test_type_handler(
            oracledb.DB_TYPE_NCLOB, oracledb.DB_TYPE_CHAR, val, val
        )

    def test_3662(self):
        "3662 - output type handler: from NCLOB to VARCHAR"
        val = "Some nclob data"
        self.__test_type_handler(
            oracledb.DB_TYPE_NCLOB, oracledb.DB_TYPE_VARCHAR, val, val
        )

    def test_3663(self):
        "3663 - output type handler: from NCLOB to LONG"
        val = "Some nclob data"
        self.__test_type_handler(
            oracledb.DB_TYPE_NCLOB, oracledb.DB_TYPE_LONG, val, val
        )

    def test_3664(self):
        "3664 - output type handler: from permanent NCLOBs to VARCHAR"
        self.__test_type_handler_lob("NCLOB", oracledb.DB_TYPE_VARCHAR)

    def test_3665(self):
        "3665 - output type handler: from permanent NCLOBs to CHAR"
        self.__test_type_handler_lob("NCLOB", oracledb.DB_TYPE_CHAR)

    def test_3666(self):
        "3666 - output type handler: from permanent NCLOBs to LONG"
        self.__test_type_handler_lob("NCLOB", oracledb.DB_TYPE_LONG)

    def test_3667(self):
        "3667 - output type handler: from NVARCHAR to VARCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_NVARCHAR, oracledb.DB_TYPE_VARCHAR, "31.5", "31.5"
        )

    def test_3668(self):
        "3668 - output type handler: from VARCHAR to NVARCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NVARCHAR, "31.5", "31.5"
        )

    def test_3669(self):
        "3669 - output type handler: from NCHAR to CHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_NCHAR, oracledb.DB_TYPE_CHAR, "31.5", "31.5"
        )

    def test_3670(self):
        "3670 - output type handler: from CHAR to NCHAR"
        self.__test_type_handler(
            oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_NCHAR, "31.5", "31.5"
        )

    def test_3671(self):
        "3671 - execute raises an error if an incorrect arraysize is used"

        def type_handler(cursor, metadata):
            return cursor.var(str)

        cursor = self.conn.cursor()
        cursor.arraysize = 100
        cursor.outputtypehandler = type_handler
        with self.assertRaisesFullCode("DPY-2016"):
            cursor.execute("select :1 from dual", [5])

    def test_3672(self):
        "3672 - execute raises an error if a var is not returned"

        def type_handler(cursor, metadata):
            return "incorrect_return"

        cursor = self.conn.cursor()
        cursor.outputtypehandler = type_handler
        with self.assertRaisesFullCode("DPY-2015"):
            cursor.execute("select :1 from dual", [5])

    def test_3673(self):
        "3673 - output type handler: from NUMBER to decimal.Decimal"
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER,
            decimal.Decimal,
            31.5,
            decimal.Decimal("31.5"),
        )
        self.__test_type_handler(
            oracledb.DB_TYPE_NUMBER, decimal.Decimal, 0, decimal.Decimal("0")
        )

    def test_3674(self):
        "3674 - use of output type handler does not affect description"

        def type_handler(cursor, metadata):
            return cursor.var(str, arraysize=cursor.arraysize)

        with self.conn.cursor() as cursor:
            cursor.execute("select user from dual")
            desc_before = cursor.description
        with self.conn.cursor() as cursor:
            cursor.outputtypehandler = type_handler
            cursor.execute("select user from dual")
            self.assertEqual(cursor.description, desc_before)

    def test_3675(self):
        "3675 - use the old signature for an output type handler"

        def type_handler(cursor, name, default_type, size, precision, scale):
            return cursor.var(str, arraysize=cursor.arraysize)

        with self.conn.cursor() as cursor:
            cursor.outputtypehandler = type_handler
            cursor.execute("select 1 from dual")
            self.assertEqual(cursor.fetchall(), [("1",)])

    def test_3676(self):
        "3676 - re-execute query with second fetch returning no rows"

        self.cursor.execute("truncate table TestTempTable")
        data = [(i + 1,) for i in range(5)]
        self.cursor.executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        self.conn.commit()

        def type_handler_1(cursor, metadata):
            return cursor.var(
                str,
                arraysize=cursor.arraysize,
                outconverter=lambda x: f"_{x}_",
            )

        def type_handler_2(cursor, metadata):
            return cursor.var(
                str,
                arraysize=cursor.arraysize,
                outconverter=lambda x: f"={x}=",
            )

        self.cursor.outputtypehandler = type_handler_1
        self.cursor.arraysize = 6
        self.cursor.prefetchrows = 6
        sql = "select IntCol from TestTempTable where rownum <= :1"
        self.cursor.execute(sql, [6])
        expected_value = [(f"_{x}_",) for x, in data]
        self.assertEqual(self.cursor.fetchall(), expected_value)

        self.cursor.outputtypehandler = type_handler_2
        self.cursor.prefetchrows = 2
        self.cursor.arraysize = 2
        self.cursor.execute(sql, [0])
        self.assertEqual(self.cursor.fetchall(), [])

    def test_3677(self):
        "3677 - output type handler: from BINARY_DOUBLE to VARCHAR"
        str_value = "36.75" if test_env.get_is_thin() else "3.675E+001"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_DOUBLE,
            oracledb.DB_TYPE_VARCHAR,
            36.75,
            str_value,
        )

    def test_3678(self):
        "3678 - output type handler: from BINARY_FLOAT to VARCHAR"
        str_value = "16.25" if test_env.get_is_thin() else "1.625E+001"
        self.__test_type_handler(
            oracledb.DB_TYPE_BINARY_FLOAT,
            oracledb.DB_TYPE_VARCHAR,
            16.25,
            str_value,
        )


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1800 - Module for testing inband notification with asyncio. No special setup
is required but the test makes use of debugging packages that are not intended
for normal use.
"""

# -----------------------------------------------------------------------------
# priv_test_1800_inband_notif_async.py
#
# Private tests for testing inband notification with asyncio. No special setup
# is required.
# -----------------------------------------------------------------------------

import unittest

import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    requires_connection = False
    setup_completed = False

    async def __perform_setup(self):
        """
        Perform setup, if needed.
        """
        if self.__class__.setup_completed:
            return
        user = test_env.get_main_user()
        async with test_env.get_admin_connection(use_async=True) as conn:
            with conn.cursor() as cursor:
                await cursor.execute(f"grant execute on dbms_tg_dbg to {user}")
        self.__class__.setup_completed = True

    async def test_ext_1800(self):
        "E1800 - test standalone connection is marked unhealthy"
        await self.__perform_setup()
        async with test_env.get_connection_async() as conn:
            self.assertEqual(conn.is_healthy(), True)
            with conn.cursor() as cursor:
                await cursor.callproc("dbms_tg_dbg.set_session_drainable")
                await cursor.execute("select user from dual")
                (user,) = await cursor.fetchone()
                self.assertEqual(user, test_env.get_main_user().upper())
            self.assertEqual(conn.is_healthy(), False)

    async def test_ext_1801(self):
        "E1801 - test pooled connection that is marked unhealthy"
        await self.__perform_setup()
        pool = test_env.get_pool_async(min=1, max=1, increment=1)
        async with pool.acquire() as conn:
            self.assertEqual(conn.is_healthy(), True)
            with conn.cursor() as cursor:
                await cursor.callproc("dbms_tg_dbg.set_session_drainable")
                info = await self.get_sid_serial(conn)
            self.assertEqual(conn.is_healthy(), False)
            with conn.cursor() as cursor:
                await cursor.execute("select user from dual")
                (user,) = await cursor.fetchone()
                self.assertEqual(user, test_env.get_main_user().upper())
        async with pool.acquire() as conn:
            self.assertEqual(conn.is_healthy(), True)
            new_info = await self.get_sid_serial(conn)
            self.assertNotEqual(new_info, info)
        await pool.close()

    async def test_ext_1802(self):
        "E1802 - test pooled connection is dropped from pool"
        await self.__perform_setup()
        pool = test_env.get_pool_async(min=1, max=1, increment=1)
        async with pool.acquire() as conn:
            self.assertEqual(conn.is_healthy(), True)
            info = await self.get_sid_serial(conn)
        async with pool.acquire() as conn:
            new_info = await self.get_sid_serial(conn)
            self.assertEqual(new_info, info)
            with conn.cursor() as cursor:
                await cursor.callproc("dbms_tg_dbg.set_session_drainable")
        async with pool.acquire() as conn:
            self.assertEqual(conn.is_healthy(), True)
            new_info = await self.get_sid_serial(conn)
            self.assertNotEqual(new_info, info)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
1000 - Module for testing top-level module methods
"""

import datetime
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def test_1000(self):
        "1000 - test DateFromTicks()"
        today = datetime.datetime.today()
        timestamp = today.timestamp()
        date = oracledb.DateFromTicks(int(timestamp))
        self.assertEqual(date, today.date())

    def test_1001(self):
        "1001 - test management of __future__ object"
        self.assertIsNone(oracledb.__future__.dummy)
        oracledb.__future__.dummy = "Unimportant"
        self.assertIsNone(oracledb.__future__.dummy)

    def test_1002(self):
        "1002 - test TimestampFromTicks()"
        timestamp = datetime.datetime.today().timestamp()
        today = datetime.datetime.fromtimestamp(timestamp)
        date = oracledb.TimestampFromTicks(timestamp)
        self.assertEqual(date, today)

    def test_1003(self):
        "1003 - test unsupported time functions"
        with self.assertRaisesFullCode("DPY-3000"):
            oracledb.Time(12, 0, 0)
        with self.assertRaisesFullCode("DPY-3000"):
            oracledb.TimeFromTicks(100)

    def test_1004(self):
        "1004 - test makedsn() with valid arguments"
        for name, value in [
            ("SID", "sid_1004"),
            ("SERVICE_NAME", "my_service_1004"),
        ]:
            host = "host_1004"
            port = 1004
            region = "US WEST"
            sharding_key = "ShardKey"
            super_sharding_key = "SuperShardKey"
            args = (
                host,
                port,
                value if name == "SID" else None,
                value if name == "SERVICE_NAME" else None,
                region,
                sharding_key,
                super_sharding_key,
            )
            result = oracledb.makedsn(*args)
            expected_value = (
                f"(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})"
                f"(PORT={port}))(CONNECT_DATA=({name}={value})"
                f"(REGION={region})(SHARDING_KEY={sharding_key})"
                f"(SUPER_SHARDING_KEY={super_sharding_key})))"
            )
            self.assertEqual(result, expected_value)

    def test_1005(self):
        "1005 - test makedsn() with invalid arguments"
        with self.assertRaisesFullCode("DPY-2020"):
            oracledb.makedsn(host="(invalid)", port=1521)
        with self.assertRaisesFullCode("DPY-2020"):
            oracledb.makedsn(host="host", port=1521, sid="(invalid)")
        with self.assertRaisesFullCode("DPY-2020"):
            oracledb.makedsn(host="host", port=1521, service_name="(invalid)")
        with self.assertRaisesFullCode("DPY-2020"):
            oracledb.makedsn(host="host", port=1521, region="(invalid)")
        with self.assertRaisesFullCode("DPY-2020"):
            oracledb.makedsn(host="host", port=1521, sharding_key="(invalid)")
        with self.assertRaisesFullCode("DPY-2020"):
            oracledb.makedsn(
                host="host", port=1521, super_sharding_key="(invalid)"
            )

    def test_1006(self):
        "1006 - test aliases match"

        # database type aliases
        self.assertIs(oracledb.BFILE, oracledb.DB_TYPE_BFILE)
        self.assertIs(oracledb.BLOB, oracledb.DB_TYPE_BLOB)
        self.assertIs(oracledb.BOOLEAN, oracledb.DB_TYPE_BOOLEAN)
        self.assertIs(oracledb.CLOB, oracledb.DB_TYPE_CLOB)
        self.assertIs(oracledb.CURSOR, oracledb.DB_TYPE_CURSOR)
        self.assertIs(oracledb.FIXED_CHAR, oracledb.DB_TYPE_CHAR)
        self.assertIs(oracledb.FIXED_NCHAR, oracledb.DB_TYPE_NCHAR)
        self.assertIs(oracledb.INTERVAL, oracledb.DB_TYPE_INTERVAL_DS)
        self.assertIs(oracledb.LONG_BINARY, oracledb.DB_TYPE_LONG_RAW)
        self.assertIs(oracledb.LONG_STRING, oracledb.DB_TYPE_LONG)
        self.assertIs(oracledb.NATIVE_INT, oracledb.DB_TYPE_BINARY_INTEGER)
        self.assertIs(oracledb.NATIVE_FLOAT, oracledb.DB_TYPE_BINARY_DOUBLE)
        self.assertIs(oracledb.NCHAR, oracledb.DB_TYPE_NVARCHAR)
        self.assertIs(oracledb.NCLOB, oracledb.DB_TYPE_NCLOB)
        self.assertIs(oracledb.OBJECT, oracledb.DB_TYPE_OBJECT)
        self.assertIs(oracledb.TIMESTAMP, oracledb.DB_TYPE_TIMESTAMP)

        # type aliases
        self.assertIs(oracledb.ObjectType, oracledb.DbObjectType)
        self.assertIs(oracledb.Object, oracledb.DbObject)
        self.assertIs(oracledb.SessionPool, oracledb.ConnectionPool)

        # authentication mode aliases
        self.assertIs(oracledb.DEFAULT_AUTH, oracledb.AUTH_MODE_DEFAULT)
        self.assertIs(oracledb.SYSASM, oracledb.AUTH_MODE_SYSASM)
        self.assertIs(oracledb.SYSBKP, oracledb.AUTH_MODE_SYSBKP)
        self.assertIs(oracledb.SYSDBA, oracledb.AUTH_MODE_SYSDBA)
        self.assertIs(oracledb.SYSDGD, oracledb.AUTH_MODE_SYSDGD)
        self.assertIs(oracledb.SYSKMT, oracledb.AUTH_MODE_SYSKMT)
        self.assertIs(oracledb.SYSOPER, oracledb.AUTH_MODE_SYSOPER)
        self.assertIs(oracledb.SYSRAC, oracledb.AUTH_MODE_SYSRAC)
        self.assertIs(oracledb.PRELIM_AUTH, oracledb.AUTH_MODE_PRELIM)

        # pool "get" mode aliases
        self.assertIs(oracledb.SPOOL_ATTRVAL_WAIT, oracledb.POOL_GETMODE_WAIT)
        self.assertIs(
            oracledb.SPOOL_ATTRVAL_NOWAIT, oracledb.POOL_GETMODE_NOWAIT
        )
        self.assertIs(
            oracledb.SPOOL_ATTRVAL_FORCEGET, oracledb.POOL_GETMODE_FORCEGET
        )
        self.assertIs(
            oracledb.SPOOL_ATTRVAL_TIMEDWAIT, oracledb.POOL_GETMODE_TIMEDWAIT
        )

        # purity aliases
        self.assertIs(oracledb.ATTR_PURITY_DEFAULT, oracledb.PURITY_DEFAULT)
        self.assertIs(oracledb.ATTR_PURITY_NEW, oracledb.PURITY_NEW)
        self.assertIs(oracledb.ATTR_PURITY_SELF, oracledb.PURITY_SELF)

        # other aliases
        self.assertIs(
            oracledb.SUBSCR_PROTO_OCI, oracledb.SUBSCR_PROTO_CALLBACK
        )
        self.assertIs(oracledb.version, oracledb.__version__)

    @unittest.skipUnless(test_env.get_is_thin(), "not relevant for thick mode")
    def test_1007(self):
        "1007 - test clientversion() fails without init_oracle_client()"
        with self.assertRaisesFullCode("DPY-2021"):
            oracledb.clientversion()

    def test_1008(self):
        "1008 - test enumeration aliases match"

        # authentication mode enumeration
        self.assertIs(oracledb.AUTH_MODE_DEFAULT, oracledb.AuthMode.DEFAULT)
        self.assertIs(oracledb.AUTH_MODE_PRELIM, oracledb.AuthMode.PRELIM)
        self.assertIs(oracledb.AUTH_MODE_SYSASM, oracledb.AuthMode.SYSASM)
        self.assertIs(oracledb.AUTH_MODE_SYSBKP, oracledb.AuthMode.SYSBKP)
        self.assertIs(oracledb.AUTH_MODE_SYSDBA, oracledb.AuthMode.SYSDBA)
        self.assertIs(oracledb.AUTH_MODE_SYSDGD, oracledb.AuthMode.SYSDGD)
        self.assertIs(oracledb.AUTH_MODE_SYSKMT, oracledb.AuthMode.SYSKMT)
        self.assertIs(oracledb.AUTH_MODE_SYSOPER, oracledb.AuthMode.SYSOPER)
        self.assertIs(oracledb.AUTH_MODE_SYSRAC, oracledb.AuthMode.SYSRAC)

        # batch operation type enumeration
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_CALL_FUNC,
            oracledb.PipelineOpType.CALL_FUNC,
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_CALL_PROC,
            oracledb.PipelineOpType.CALL_PROC,
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_COMMIT, oracledb.PipelineOpType.COMMIT
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_EXECUTE, oracledb.PipelineOpType.EXECUTE
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_EXECUTE_MANY,
            oracledb.PipelineOpType.EXECUTE_MANY,
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_FETCH_ALL,
            oracledb.PipelineOpType.FETCH_ALL,
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_FETCH_MANY,
            oracledb.PipelineOpType.FETCH_MANY,
        )
        self.assertIs(
            oracledb.PIPELINE_OP_TYPE_FETCH_ONE,
            oracledb.PipelineOpType.FETCH_ONE,
        )

        # pool "get" mode enumeration
        self.assertIs(
            oracledb.POOL_GETMODE_FORCEGET, oracledb.PoolGetMode.FORCEGET
        )
        self.assertIs(oracledb.POOL_GETMODE_WAIT, oracledb.PoolGetMode.WAIT)
        self.assertIs(
            oracledb.POOL_GETMODE_NOWAIT, oracledb.PoolGetMode.NOWAIT
        )
        self.assertIs(
            oracledb.POOL_GETMODE_TIMEDWAIT, oracledb.PoolGetMode.TIMEDWAIT
        )

        # purity enumeration
        self.assertIs(oracledb.PURITY_DEFAULT, oracledb.Purity.DEFAULT)
        self.assertIs(oracledb.PURITY_NEW, oracledb.Purity.NEW)
        self.assertIs(oracledb.PURITY_SELF, oracledb.Purity.SELF)

        # vector format enumeration
        self.assertIs(
            oracledb.VECTOR_FORMAT_BINARY, oracledb.VectorFormat.BINARY
        )
        self.assertIs(
            oracledb.VECTOR_FORMAT_FLOAT32, oracledb.VectorFormat.FLOAT32
        )
        self.assertIs(
            oracledb.VECTOR_FORMAT_FLOAT64, oracledb.VectorFormat.FLOAT64
        )
        self.assertIs(oracledb.VECTOR_FORMAT_INT8, oracledb.VectorFormat.INT8)

    def test_1009(self):
        "1009 - test enable_thin_mode()"
        if test_env.get_is_thin():
            oracledb.enable_thin_mode()
            with self.assertRaisesFullCode("DPY-2019"):
                oracledb.init_oracle_client()
        else:
            with self.assertRaisesFullCode("DPY-2053"):
                oracledb.enable_thin_mode()


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# dataframe.py
#
# Implement DataFrame class as documented in the standard
# https://data-apis.org/dataframe-protocol/latest/API.html
# -----------------------------------------------------------------------------

from typing import Any, Dict, Iterable, List, Optional, Sequence

from .column import OracleColumn

from .protocol import DataFrame


class OracleDataFrame(DataFrame):
    """
    OracleDataFrame is an implementation of the DataFrame Interchange Protocol.
    It provides an interface for exchanging tabular data between different data
    frame libraries (e.g. pandas, pyarrow, polars).
    """

    def __init__(
        self,
        oracle_arrow_arrays: List,
        allow_copy: bool = True,
    ):
        self._cols = []
        self._cols_map = {}
        self._rows = None
        self._arrays = oracle_arrow_arrays
        for ora_arrow_array in oracle_arrow_arrays:
            column = OracleColumn(ora_arrow_array=ora_arrow_array)
            self._rows = column.size()
            self._cols.append(column)
            self._cols_map[ora_arrow_array.name] = column
        self.allow_copy = allow_copy

    def __dataframe__(
        self,
        nan_as_null: bool = False,  # noqa: FBT001
        allow_copy: bool = True,  # noqa: FBT001
    ) -> DataFrame:
        """
        Returns a data frame adhering to the DataFrame Interchange protocol.
        """
        return self

    def get_chunks(
        self, n_chunks: Optional[int] = None
    ) -> Iterable[DataFrame]:
        """
        Returns an iterator for each of the chunks in the data frame. Since
        there is currently only one chunk, this simply returns itself.
        """
        yield self

    def column_arrays(self) -> List:
        """
        Returns a list of the Arrow arrays corresponding to each column in the
        data frame.
        """
        return self._arrays

    def column_names(self) -> List[str]:
        """
        Returns a list of the names of the columns in the data frame.
        """
        return list(self._cols_map.keys())

    def get_column(self, i: int) -> OracleColumn:
        """
        Returns a column from the data frame given its zero-based index. If the
        index is out of range, an IndexError exception is raised.
        """
        if i < 0 or i >= self.num_columns():
            raise IndexError(
                f"Column index {i} is out of bounds for "
                f"DataFrame with {self.num_columns()} columns"
            )
        return self._cols[i]

    def get_column_by_name(self, name: str) -> OracleColumn:
        """
        Returns a column from the data frame given the name of the column. If
        the column name is not found, a KeyError exception is raised.
        """
        if name not in self._cols_map:
            raise KeyError(f"Column {name} not found in DataFrame")
        return self._cols_map[name]

    def get_columns(self) -> List[OracleColumn]:
        """
        Returns a list of all of the columns in the data frame.
        """
        return self._cols

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Returns metadata for the data frame. Currently this returns
        information about the number of columns (num_columns), number of rows
        (num_rows) and number of chunks (num_chunks).
        """
        return {
            "num_columns": self.num_columns(),
            "num_rows": self.num_rows(),
            "num_chunks": self.num_chunks(),
        }

    def num_chunks(self) -> int:
        """
        Returns the number of chunks (contiguous memory blocks) in the data
        frame. Currently this always returns 1.
        """
        return 1

    def num_columns(self) -> int:
        """
        Returns the number of columns in the data frame.
        """
        return len(self._cols)

    def num_rows(self) -> int:
        """
        Returns the number of rows in the data frame.
        """
        return self._rows

    def select_columns(self, indices: Sequence[int]) -> "DataFrame":
        """
        Create a new DataFrame by selecting a subset of columns by index.
        """
        raise NotImplementedError()

    def select_columns_by_name(self, names: Sequence[str]) -> "DataFrame":
        """
        Create a new DataFrame by selecting a subset of columns by name.
        """
        raise NotImplementedError()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# lob_op.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for performing LOB operations
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class LobOpMessage(Message):
    cdef:
        uint32_t operation
        BaseThinLobImpl source_lob_impl
        uint64_t source_offset
        uint64_t dest_offset
        uint32_t dest_length
        int64_t amount
        bint send_amount
        bint bool_flag
        object data

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization.
        """
        self.function_code = TNS_FUNC_LOB_OP

    cdef int _process_message(self, ReadBuffer buf,
                              uint8_t message_type) except -1:
        cdef:
            const char* encoding
            const char_type *ptr
            ssize_t num_bytes
        if message_type == TNS_MSG_TYPE_LOB_DATA:
            buf.read_raw_bytes_and_length(&ptr, &num_bytes)
            if self.source_lob_impl.dbtype._ora_type_num in \
                    (ORA_TYPE_NUM_BLOB, ORA_TYPE_NUM_BFILE):
                self.data = ptr[:num_bytes]
            else:
                encoding = self.source_lob_impl._get_encoding()
                self.data = ptr[:num_bytes].decode(encoding)
        else:
            Message._process_message(self, buf, message_type)

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        cdef:
            cdef const char_type *ptr
            ssize_t num_bytes
            uint8_t temp8
        if self.source_lob_impl is not None:
            num_bytes = len(self.source_lob_impl._locator)
            ptr = buf.read_raw_bytes(num_bytes)
            self.source_lob_impl._locator = ptr[:num_bytes]
        if self.operation == TNS_LOB_OP_CREATE_TEMP:
            buf.skip_ub2()                  # skip character set
            buf.skip_ub1()                  # skip trailing flags
        elif self.send_amount:
            buf.read_sb8(&self.amount)
        if self.operation in (TNS_LOB_OP_IS_OPEN,
                              TNS_LOB_OP_FILE_EXISTS,
                              TNS_LOB_OP_FILE_ISOPEN):
            buf.read_ub1(&temp8)
            self.bool_flag = temp8 > 0

    cdef int _write_message(self, WriteBuffer buf) except -1:
        cdef int i
        self._write_function_code(buf)
        if self.source_lob_impl is None:
            buf.write_uint8(0)              # source pointer
            buf.write_ub4(0)                # source length
        else:
            buf.write_uint8(1)              # source pointer
            buf.write_ub4(len(self.source_lob_impl._locator))
        buf.write_uint8(0)                  # dest pointer
        buf.write_ub4(self.dest_length)
        buf.write_ub4(0)                    # short source offset
        buf.write_ub4(0)                    # short dest offset
        if self.operation == TNS_LOB_OP_CREATE_TEMP:
            buf.write_uint8(1)              # pointer (character set)
        else:
            buf.write_uint8(0)              # pointer (character set)
        buf.write_uint8(0)                  # pointer (short amount)
        if self.operation in (TNS_LOB_OP_CREATE_TEMP,
                              TNS_LOB_OP_IS_OPEN,
                              TNS_LOB_OP_FILE_EXISTS,
                              TNS_LOB_OP_FILE_ISOPEN):
            buf.write_uint8(1)              # pointer (NULL LOB)
        else:
            buf.write_uint8(0)              # pointer (NULL LOB)
        buf.write_ub4(self.operation)
        buf.write_uint8(0)                  # pointer (SCN array)
        buf.write_uint8(0)                  # SCN array length
        buf.write_ub8(self.source_offset)
        buf.write_ub8(self.dest_offset)
        if self.send_amount:
            buf.write_uint8(1)              # pointer (amount)
        else:
            buf.write_uint8(0)              # pointer (amount)
        for i in range(3):                  # array LOB (not used)
            buf.write_uint16be(0)
        if self.source_lob_impl is not None:
            buf.write_bytes(self.source_lob_impl._locator)
        if self.operation == TNS_LOB_OP_CREATE_TEMP:
            if self.source_lob_impl.dbtype._csfrm == CS_FORM_NCHAR:
                buf._caps._check_ncharset_id()
                buf.write_ub4(TNS_CHARSET_UTF16)
            else:
                buf.write_ub4(TNS_CHARSET_UTF8)
        if self.data is not None:
            buf.write_uint8(TNS_MSG_TYPE_LOB_DATA)
            buf.write_bytes_with_length(self.data)
        if self.send_amount:
            buf.write_ub8(self.amount)      # LOB amount
